@@ -95,6 +95,30 @@ from .interface import (
     UpdatablePenalizationFunction,
     VectorizedAcquisitionFunctionBuilder,
 )
+from .rule import (
+    AcquisitionRule,
+    AsynchronousGreedy,
+    AsynchronousOptimization,
+    BatchHypervolumeSharpeRatioIndicator,
+    BatchTrustRegion,
+    BatchTrustRegionBox,
+    BatchTrustRegionProduct,
+    DiscreteThompsonSampling,
+    EfficientGlobalOptimization,
+    FixedPointTrustRegionDiscrete,
+    HypercubeTrustRegion,
+    LocalDatasetsAcquisitionRule,
+    RandomSampling,
+    SingleObjectiveTrustRegionBox,
+    SingleObjectiveTrustRegionDiscrete,
+    TREGOBox,
+    TURBOBox,
+    UpdatableSearchSpace,
+    UpdatableTrustRegion,
+    UpdatableTrustRegionBox,
+    UpdatableTrustRegionDiscrete,
+    UpdatableTrustRegionProduct,
+)
 from .sampler import (
     ExactThompsonSampler,
     GumbelSampler,
