@@ -15,17 +15,17 @@ from __future__ import annotations
 
 from abc import abstractmethod
 from collections.abc import Mapping, Sequence
-from typing import Optional
+from typing import Callable, Optional
 
 import tensorflow as tf
 
 from ..data import Dataset
-from ..models import ProbabilisticModel
+from ..models import ProbabilisticModelType
 from ..types import Tag, TensorType
 from .interface import AcquisitionFunction, AcquisitionFunctionBuilder
 
 
-class Reducer(AcquisitionFunctionBuilder[ProbabilisticModel]):
+class Reducer(AcquisitionFunctionBuilder[ProbabilisticModelType]):
     r"""
     A :class:`Reducer` builds an :const:`~trieste.acquisition.AcquisitionFunction` whose output is
     calculated from the outputs of a number of other
@@ -33,7 +33,7 @@ class Reducer(AcquisitionFunctionBuilder[ProbabilisticModel]):
     by the method :meth:`_reduce`.
     """
 
-    def __init__(self, *builders: AcquisitionFunctionBuilder[ProbabilisticModel]):
+    def __init__(self, *builders: AcquisitionFunctionBuilder[ProbabilisticModelType]):
         r"""
         :param \*builders: Acquisition function builders. At least one must be provided.
         :raise `~tf.errors.InvalidArgumentError`: If no builders are specified.
@@ -44,12 +44,14 @@ class Reducer(AcquisitionFunctionBuilder[ProbabilisticModel]):
 
         self._acquisitions = builders
 
-    def _repr_builders(self) -> str:
-        return ", ".join(map(repr, self._acquisitions))
+    def __repr__(self) -> str:
+        """"""
+        builders = ", ".join(map(repr, self._acquisitions))
+        return f"{self.__class__.__name__}({builders})"
 
     def prepare_acquisition_function(
         self,
-        models: Mapping[Tag, ProbabilisticModel],
+        models: Mapping[Tag, ProbabilisticModelType],
         datasets: Optional[Mapping[Tag, Dataset]] = None,
     ) -> AcquisitionFunction:
         r"""
@@ -75,7 +77,7 @@ class Reducer(AcquisitionFunctionBuilder[ProbabilisticModel]):
     def update_acquisition_function(
         self,
         function: AcquisitionFunction,
-        models: Mapping[Tag, ProbabilisticModel],
+        models: Mapping[Tag, ProbabilisticModelType],
         datasets: Optional[Mapping[Tag, Dataset]] = None,
     ) -> AcquisitionFunction:
         """
@@ -94,7 +96,7 @@ class Reducer(AcquisitionFunctionBuilder[ProbabilisticModel]):
         return evaluate_acquisition_function_fn
 
     @property
-    def acquisitions(self) -> Sequence[AcquisitionFunctionBuilder[ProbabilisticModel]]:
+    def acquisitions(self) -> Sequence[AcquisitionFunctionBuilder[ProbabilisticModelType]]:
         """The acquisition function builders specified at class initialisation."""
         return self._acquisitions
 
@@ -112,16 +114,12 @@ class Reducer(AcquisitionFunctionBuilder[ProbabilisticModel]):
         raise NotImplementedError()
 
 
-class Sum(Reducer):
+class Sum(Reducer[ProbabilisticModelType]):
     """
     :class:`Reducer` whose resulting acquisition function returns the element-wise sum of the
     outputs of constituent acquisition functions.
     """
 
-    def __repr__(self) -> str:
-        """"""
-        return f"Sum({self._repr_builders()})"
-
     def _reduce(self, inputs: Sequence[TensorType]) -> TensorType:
         """
         :param inputs: The outputs of each acquisition function.
@@ -130,19 +128,43 @@ class Sum(Reducer):
         return tf.add_n(inputs)
 
 
-class Product(Reducer):
+class Product(Reducer[ProbabilisticModelType]):
     """
     :class:`Reducer` whose resulting acquisition function returns the element-wise product of the
     outputs of constituent acquisition functions.
     """
 
-    def __repr__(self) -> str:
-        """"""
-        return f"Product({self._repr_builders()})"
-
     def _reduce(self, inputs: Sequence[TensorType]) -> TensorType:
         """
         :param inputs: The outputs of each acquisition function.
         :return: The element-wise product of the ``inputs``.
         """
         return tf.reduce_prod(inputs, axis=0)
+
+
+class Map(Reducer[ProbabilisticModelType]):
+    """
+    :class:`Reducer` that accepts just one acquisition function builder and applies a
+    given function to its output. For example ``Map(lambda x: -x, builder)`` would generate
+    an acquisition function that returns the negative of the output of ``builder``.
+    """
+
+    def __init__(
+        self,
+        map_fn: Callable[[TensorType], TensorType],
+        builder: AcquisitionFunctionBuilder[ProbabilisticModelType],
+    ):
+        """
+        :param map_fn: Function to apply.
+        :param builder: Acquisition function builder.
+        """
+        super().__init__(builder)
+        self._map_fn = map_fn
+
+    def _reduce(self, inputs: Sequence[TensorType]) -> TensorType:
+        """
+        :param inputs: The outputs of the acquisition function.
+        :return: The result of applying the map function to ``inputs``.
+        """
+        tf.debugging.assert_equal(len(inputs), 1)
+        return self._map_fn(inputs[0])
