@@ -29,6 +29,7 @@ from ...models.interfaces import (
     HasTrajectorySampler,
     SupportsCovarianceWithTopFidelity,
     SupportsGetObservationNoise,
+    SupportsPredictY,
 )
 from ...space import SearchSpace
 from ...types import TensorType
@@ -204,7 +205,7 @@ class min_value_entropy_search(AcquisitionFunctionClass):
             fsd, CLAMP_LB, fmean.dtype.max
         )  # clip below to improve numerical stability
 
-        normal = tfp.distributions.Normal(tf.cast(0, fmean.dtype), tf.cast(1, fmean.dtype))
+        normal = tfp.distributions.Normal(tf.constant(0, fmean.dtype), tf.constant(1, fmean.dtype))
         gamma = (tf.squeeze(self._samples) - fmean) / fsd
 
         log_minus_cdf = normal.log_cdf(-gamma)
@@ -220,8 +221,6 @@ class SupportsCovarianceObservationNoise(
 ):
     """A model that supports both covariance_between_points and get_observation_noise."""
 
-    pass
-
 
 class SupportsCovarianceObservationNoiseTrajectory(
     HasTrajectorySampler, SupportsCovarianceObservationNoise, Protocol
@@ -336,7 +335,7 @@ class GIBBON(SingleModelGreedyAcquisitionBuilder[GIBBONModelType]):
         if not isinstance(model, SupportsCovarianceObservationNoise):
             raise NotImplementedError(
                 f"GIBBON only works with models that support "
-                f"covariance_between_points and get_observation_noise; received {model.__repr__()}"
+                f"covariance_between_points and get_observation_noise; received {model!r}"
             )
 
         tf.debugging.Assert(dataset is not None, [tf.constant([])])
@@ -497,7 +496,7 @@ class gibbon_quality_term(AcquisitionFunctionClass):
         )  # clip below to improve numerical stability
         gamma = (tf.squeeze(self._samples) - fmean) / fsd
 
-        normal = tfp.distributions.Normal(tf.cast(0, fmean.dtype), tf.cast(1, fmean.dtype))
+        normal = tfp.distributions.Normal(tf.constant(0, fmean.dtype), tf.constant(1, fmean.dtype))
         log_minus_cdf = normal.log_cdf(-gamma)
         ratio = tf.math.exp(normal.log_prob(gamma) - log_minus_cdf)
         inner_log = 1 + rho_squared * ratio * (gamma - ratio)
@@ -623,10 +622,17 @@ class gibbon_repulsion_term(UpdatablePenalizationFunction):
         return repulsion_weight * repulsion
 
 
+@runtime_checkable
+class SupportsCovarianceWithTopFidelityPredictY(
+    SupportsCovarianceWithTopFidelity, SupportsPredictY, Protocol
+):
+    """A model that is both multifidelity and supports predict_y."""
+
+
 MUMBOModelType = TypeVar(
-    "MUMBOModelType", bound=SupportsCovarianceWithTopFidelity, contravariant=True
+    "MUMBOModelType", bound=SupportsCovarianceWithTopFidelityPredictY, contravariant=True
 )
-""" Type variable bound to :class:`~trieste.models.SupportsCovarianceWithTopFidelity`. """
+""" Type variable bound to :class:`~trieste.models.SupportsCovarianceWithTopFidelityPredictY`. """
 
 
 class MUMBO(MinValueEntropySearch[MUMBOModelType]):
@@ -645,7 +651,7 @@ class MUMBO(MinValueEntropySearch[MUMBOModelType]):
 
     @overload
     def __init__(
-        self: "MUMBO[SupportsCovarianceWithTopFidelity]",
+        self: "MUMBO[SupportsCovarianceWithTopFidelityPredictY]",
         search_space: SearchSpace,
         num_samples: int = 5,
         grid_size: int = 1000,
@@ -779,7 +785,7 @@ class mumbo(AcquisitionFunctionClass):
         rho_squared = (cov**2) / (fvar * yvar)
         rho_squared = tf.clip_by_value(rho_squared, 0.0, 1.0)
 
-        normal = tfp.distributions.Normal(tf.cast(0, fmean.dtype), tf.cast(1, fmean.dtype))
+        normal = tfp.distributions.Normal(tf.constant(0, fmean.dtype), tf.constant(1, fmean.dtype))
         gamma = (tf.squeeze(self._samples) - fmean) / fsd
         log_minus_cdf = normal.log_cdf(-gamma)
         ratio = tf.math.exp(normal.log_prob(gamma) - log_minus_cdf)
