@@ -21,6 +21,7 @@ from typing import Callable, Mapping, Optional, cast
 
 import tensorflow as tf
 import tensorflow_probability as tfp
+from check_shapes import check_shapes
 
 from ...data import Dataset
 from ...models import ProbabilisticModel, ReparametrizationSampler
@@ -212,11 +213,11 @@ class expected_improvement(AcquisitionFunctionClass):
         self._eta.assign(eta)
 
     @tf.function
+    @check_shapes(
+        "x: [N..., 1, D] # This acquisition function only supports batch sizes of one",
+        "return: [N..., L]",
+    )
     def __call__(self, x: TensorType) -> TensorType:
-        tf.debugging.assert_shapes(
-            [(x, [..., 1, None])],
-            message="This acquisition function only supports batch sizes of one.",
-        )
         mean, variance = self._model.predict(tf.squeeze(x, -2))
         normal = tfp.distributions.Normal(mean, tf.sqrt(variance))
         return (self._eta - mean) * normal.cdf(self._eta) + variance * normal.prob(self._eta)
@@ -248,7 +249,7 @@ class AugmentedExpectedImprovement(SingleModelAcquisitionBuilder[SupportsGetObse
         if not isinstance(model, SupportsGetObservationNoise):
             raise NotImplementedError(
                 f"AugmentedExpectedImprovement only works with models that support "
-                f"get_observation_noise; received {model.__repr__()}"
+                f"get_observation_noise; received {model!r}"
             )
         tf.debugging.Assert(dataset is not None, [tf.constant([])])
         dataset = cast(Dataset, dataset)
@@ -317,14 +318,11 @@ class augmented_expected_improvement(AcquisitionFunctionClass):
         )
         mean, variance = self._model.predict(tf.squeeze(x, -2))
         normal = tfp.distributions.Normal(mean, tf.sqrt(variance))
-        expected_improvement = (self._eta - mean) * normal.cdf(self._eta) + variance * normal.prob(
-            self._eta
-        )
-
+        ei = (self._eta - mean) * normal.cdf(self._eta) + variance * normal.prob(self._eta)
         augmentation = 1 - (tf.math.sqrt(self._noise_variance)) / (
             tf.math.sqrt(self._noise_variance + variance)
         )
-        return expected_improvement * augmentation
+        return ei * augmentation
 
 
 class NegativeLowerConfidenceBound(SingleModelAcquisitionBuilder[ProbabilisticModel]):
@@ -597,7 +595,7 @@ def fast_constraints_feasibility(
     def acquisition(x: TensorType) -> TensorType:
         if smoothing_function is None:
             _smoothing_function = tfp.distributions.Normal(
-                tf.cast(0.0, x.dtype), tf.cast(1e-3, x.dtype)
+                tf.constant(0.0, x.dtype), tf.constant(1e-3, x.dtype)
             ).cdf
         else:
             _smoothing_function = smoothing_function
@@ -639,8 +637,8 @@ class ExpectedConstrainedImprovement(AcquisitionFunctionBuilder[ProbabilisticMod
             tf.debugging.assert_less_equal(float(min_feasibility_probability), 1.0)
         else:
             dtype = min_feasibility_probability.dtype
-            tf.debugging.assert_greater_equal(min_feasibility_probability, tf.cast(0, dtype))
-            tf.debugging.assert_less_equal(min_feasibility_probability, tf.cast(1, dtype))
+            tf.debugging.assert_greater_equal(min_feasibility_probability, tf.constant(0, dtype))
+            tf.debugging.assert_less_equal(min_feasibility_probability, tf.constant(1, dtype))
 
         self._objective_tag = objective_tag
         self._constraint_builder = constraint_builder
@@ -827,7 +825,7 @@ class MonteCarloExpectedImprovement(SingleModelAcquisitionBuilder[HasReparamSamp
         if not isinstance(model, HasReparamSampler):
             raise ValueError(
                 f"MonteCarloExpectedImprovement only supports models with a reparam_sampler method;"
-                f"received {model.__repr__()}"
+                f"received {model!r}"
             )
 
         sampler = model.reparam_sampler(self._sample_size)
@@ -971,7 +969,7 @@ class MonteCarloAugmentedExpectedImprovement(
             raise ValueError(
                 f"MonteCarloAugmentedExpectedImprovement only supports models with a "
                 f"reparam_sampler method and that support observation noise; received "
-                f"{model.__repr__()}."
+                f"{model!r}."
             )
 
         sampler = model.reparam_sampler(self._sample_size)
@@ -1166,7 +1164,7 @@ class batch_monte_carlo_expected_improvement(AcquisitionFunctionClass):
         if not isinstance(model, HasReparamSampler):
             raise ValueError(
                 f"The batch Monte-Carlo expected improvement acquisition function only supports "
-                f"models that implement a reparam_sampler method; received {model.__repr__()}"
+                f"models that implement a reparam_sampler method; received {model!r}"
             )
 
         sampler = model.reparam_sampler(self._sample_size)
@@ -1218,7 +1216,7 @@ class BatchExpectedImprovement(SingleModelAcquisitionBuilder[ProbabilisticModel]
     def __repr__(self) -> str:
         """"""
 
-        return f"BatchExpectedImprovement({self._sample_size!r}, " f"jitter={self._jitter!r})"
+        return f"BatchExpectedImprovement({self._sample_size!r}, jitter={self._jitter!r})"
 
     def prepare_acquisition_function(
         self,
@@ -1742,9 +1740,9 @@ class batch_expected_improvement(AcquisitionFunctionClass):
         )
 
         # Compute outer sum
-        expected_improvement = tf.reduce_sum(mean_T_term + sum_term, axis=1)
+        ei = tf.reduce_sum(mean_T_term + sum_term, axis=1)
 
-        return expected_improvement
+        return ei
 
     @tf.function
     def __call__(self, x: TensorType) -> TensorType:
@@ -1898,7 +1896,7 @@ class multiple_optimism_lower_confidence_bound(AcquisitionFunctionClass):
 
         if not self._initialized:
             normal = tfp.distributions.Normal(
-                tf.cast(0.0, dtype=x.dtype), tf.cast(1.0, dtype=x.dtype)
+                tf.constant(0.0, dtype=x.dtype), tf.constant(1.0, dtype=x.dtype)
             )
             spread = 0.5 + 0.5 * tf.range(1, batch_size + 1, dtype=x.dtype) / (
                 tf.cast(batch_size, dtype=x.dtype) + 1.0
