@@ -21,15 +21,19 @@ from typing import Callable, Dict, Mapping, Optional, Union, cast
 import gpflow
 import tensorflow as tf
 import tensorflow_probability as tfp
+from check_shapes import check_shapes
 from typing_extensions import Protocol, runtime_checkable
 
 from ...data import Dataset
 from ...models import FastUpdateModel, ModelStack, ProbabilisticModel
 from ...models.interfaces import (
     PredictJointModelStack,
+    PredictJointPredictYModelStack,
+    PredictYModelStack,
     SupportsGetKernel,
     SupportsGetObservationNoise,
     SupportsPredictJoint,
+    SupportsPredictY,
 )
 from ...observer import OBJECTIVE
 from ...space import SearchSpace
@@ -343,7 +347,7 @@ class soft_local_penalizer(local_penalizer):
         )
         standardised_distances = (pairwise_distances - self._radius) / self._scale
 
-        normal = tfp.distributions.Normal(tf.cast(0, x.dtype), tf.cast(1, x.dtype))
+        normal = tfp.distributions.Normal(tf.constant(0, x.dtype), tf.constant(1, x.dtype))
         penalization = normal.cdf(standardised_distances)
         return tf.reduce_prod(penalization, axis=-1)
 
@@ -384,21 +388,24 @@ class hard_local_penalizer(local_penalizer):
 
 @runtime_checkable
 class FantasizerModelType(
-    FastUpdateModel, SupportsPredictJoint, SupportsGetKernel, SupportsGetObservationNoise, Protocol
+    FastUpdateModel,
+    SupportsPredictJoint,
+    SupportsPredictY,
+    SupportsGetKernel,
+    SupportsGetObservationNoise,
+    Protocol,
 ):
     """The model requirements for the Fantasizer acquisition function."""
 
-    pass
 
-
-class FantasizerModelStack(PredictJointModelStack, ModelStack[FantasizerModelType]):
+class FantasizerModelStack(
+    PredictJointModelStack, PredictYModelStack, ModelStack[FantasizerModelType]
+):
     """
     A stack of models :class:`FantasizerModelType` models. Note that this delegates predict_joint
-    but none of the other methods.
+    and predict_y but none of the other methods.
     """
 
-    pass
-
 
 FantasizerModelOrStack = Union[FantasizerModelType, FantasizerModelStack]
 
@@ -544,7 +551,7 @@ class Fantasizer(GreedyAcquisitionFunctionBuilder[FantasizerModelOrStack]):
                 raise NotImplementedError(
                     f"Fantasizer only works with FastUpdateModel models that also support "
                     f"predict_joint, get_kernel and get_observation_noise, or with "
-                    f"ModelStack stacks of such models; received {model.__repr__()}"
+                    f"ModelStack stacks of such models; received {model!r}"
                 )
         if pending_points is None:
             return self._update_base_acquisition_function(models, datasets)
@@ -595,16 +602,14 @@ def _generate_fantasized_data(
     elif fantasize_method == "sample":
         fantasized_obs = model.sample(pending_points, num_samples=1)[0]
     else:
-        raise NotImplementedError(
-            f"fantasize_method must be KB or sample, " f"received {model.__repr__()}"
-        )
+        raise NotImplementedError(f"fantasize_method must be KB or sample, received {model!r}")
 
     return Dataset(pending_points, fantasized_obs)
 
 
 def _generate_fantasized_model(
     model: FantasizerModelOrStack, fantasized_data: Dataset
-) -> _fantasized_model | PredictJointModelStack:
+) -> _fantasized_model | PredictJointPredictYModelStack:
     if isinstance(model, ModelStack):
         observations = tf.split(fantasized_data.observations, model._event_sizes, axis=-1)
         fmods = []
@@ -615,12 +620,14 @@ def _generate_fantasized_model(
                     event_size,
                 )
             )
-        return PredictJointModelStack(*fmods)
+        return PredictJointPredictYModelStack(*fmods)
     else:
         return _fantasized_model(model, fantasized_data)
 
 
-class _fantasized_model(SupportsPredictJoint, SupportsGetKernel, SupportsGetObservationNoise):
+class _fantasized_model(
+    SupportsPredictJoint, SupportsGetKernel, SupportsGetObservationNoise, SupportsPredictY
+):
     """
     Creates a new model from an existing one and additional data.
     This new model posterior is conditioned on both current model data and the additional one.
@@ -652,6 +659,11 @@ class _fantasized_model(SupportsPredictJoint, SupportsGetKernel, SupportsGetObse
         self._fantasized_query_points.assign(fantasized_data.query_points)
         self._fantasized_observations.assign(fantasized_data.observations)
 
+    @check_shapes(
+        "query_points: [batch..., N, D]",
+        "return[0]: [batch..., ..., N, L]",
+        "return[1]: [batch..., ..., N, L]",
+    )
     def predict(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
         """
         This function wraps conditional_predict_f. It cannot directly call
@@ -659,7 +671,7 @@ class _fantasized_model(SupportsPredictJoint, SupportsGetKernel, SupportsGetObse
         We use map_fn to allow leading dimensions for query_points.
 
         :param query_points: shape [...*, N, d]
-        :return: mean, shape [...*, ..., N, L] and cov, shape [...*, ..., L, N],
+        :return: mean, shape [...*, ..., N, L] and cov, shape [...*, ..., N, L],
             where ... are the leading dimensions of fantasized_data
         """
 
@@ -671,6 +683,11 @@ class _fantasized_model(SupportsPredictJoint, SupportsGetKernel, SupportsGetObse
 
         return _broadcast_predict(query_points, fun)
 
+    @check_shapes(
+        "query_points: [batch..., N, D]",
+        "return[0]: [batch..., ..., N, L]",
+        "return[1]: [batch..., ..., L, N, N]",
+    )
     def predict_joint(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
         """
         This function wraps conditional_predict_joint. It cannot directly call
@@ -690,6 +707,10 @@ class _fantasized_model(SupportsPredictJoint, SupportsGetKernel, SupportsGetObse
 
         return _broadcast_predict(query_points, fun)
 
+    @check_shapes(
+        "query_points: [batch..., N, D]",
+        "return: [batch..., ..., S, N, L]",
+    )
     def sample(self, query_points: TensorType, num_samples: int) -> TensorType:
         """
         This function wraps conditional_predict_f_sample. It cannot directly call
@@ -716,6 +737,11 @@ class _fantasized_model(SupportsPredictJoint, SupportsGetKernel, SupportsGetObse
         )  # [B, ..., S, L]
         return _restore_leading_dim(samples, leading_dim)
 
+    @check_shapes(
+        "query_points: [broadcast batch..., N, D]",
+        "return[0]: [batch..., ..., N, L]",
+        "return[1]: [batch..., ..., N, L]",
+    )
     def predict_y(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
         """
         This function wraps conditional_predict_y. It cannot directly call
@@ -723,7 +749,7 @@ class _fantasized_model(SupportsPredictJoint, SupportsGetKernel, SupportsGetObse
         We use tf.map_fn to allow leading dimensions for query_points.
 
         :param query_points: shape [...*, N, D]
-        :return: mean, shape [...*, ..., N, L] and var, shape [...*, ..., L, N],
+        :return: mean, shape [...*, ..., N, L] and var, shape [...*, ..., N, L],
             where ... are the leading dimensions of fantasized_data
         """
 
