@@ -19,21 +19,25 @@ This module contains functionality for optimizing
 
 from __future__ import annotations
 
-from typing import Any, Callable, List, Optional, Sequence, Tuple, Union, cast
+from itertools import chain
+from typing import Any, Callable, Iterable, List, Optional, Sequence, Tuple, Union, cast
 
 import greenlet as gr
 import numpy as np
 import scipy.optimize as spo
 import tensorflow as tf
 import tensorflow_probability as tfp
+from check_shapes import check_shapes
 
 from .. import logging
 from ..space import (
     Box,
+    CollectionSearchSpace,
     Constraint,
     DiscreteSearchSpace,
     SearchSpace,
     SearchSpaceType,
+    TaggedMultiSearchSpace,
     TaggedProductSearchSpace,
 )
 from ..types import TensorType
@@ -100,9 +104,10 @@ def automatic_optimizer_selector(
     if isinstance(space, DiscreteSearchSpace):
         return optimize_discrete(space, target_func)
 
-    elif isinstance(space, (Box, TaggedProductSearchSpace)):
-        num_samples = tf.maximum(NUM_SAMPLES_MIN, NUM_SAMPLES_DIM * tf.shape(space.lower)[-1])
-        num_runs = NUM_RUNS_DIM * tf.shape(space.lower)[-1]
+    elif isinstance(space, (Box, CollectionSearchSpace)):
+        space_dim = space.dimension
+        num_samples = tf.maximum(NUM_SAMPLES_MIN, NUM_SAMPLES_DIM * space_dim)
+        num_runs = NUM_RUNS_DIM * space_dim
         return generate_continuous_optimizer(
             num_initial_samples=num_samples,
             num_optimization_runs=num_runs,
@@ -165,15 +170,155 @@ def optimize_discrete(
     return _get_max_discrete_points(points, target_func)
 
 
+InitialPointSampler = Callable[[SearchSpace], Iterable[TensorType]]
+"""
+Type alias for a function that returns initial point candidates for an optimization.
+Candidates are returned in one or more batches, and each batch should have the shape [N, D],
+even when N=1.
+
+For simplicity and memory usage, it is recommended to define these as generators. For example,
+the following initial point sampler returns both a set of pre-optimized points and 50,000
+random samples:
+
+    def sampler(space: SearchSpace) -> Iterable[TensorType]:
+        yield pre_optimized_points
+        yield space.sample(50_000)
+
+While the following does the same but groups the random samples into batches of size 1,000
+to conserve memory:
+
+    def sampler(space: SearchSpace) -> Iterable[TensorType]:
+        yield pre_optimized_points
+        yield from sample_from_space(50_000, batch_size=1_000)(space)
+"""
+
+
+def sample_from_space(num_samples: int, batch_size: Optional[int] = None) -> InitialPointSampler:
+    """
+    An initial point sampler that returns `num_samples` points. If `batch_size` is specified,
+    then these are returned in batches of that size, to preserve memory usage.
+    """
+    if num_samples <= 0:
+        raise ValueError(f"num_samples must be positive, got {num_samples}")
+
+    if isinstance(batch_size, int) and batch_size <= 0:
+        raise ValueError(f"batch_size must be positive, got {batch_size}")
+
+    batch_size_int = batch_size or num_samples
+
+    def sampler(space: SearchSpace) -> Iterable[TensorType]:
+        for offset in range(0, num_samples, batch_size_int):
+            yield space.sample(min(num_samples - offset, batch_size_int))
+
+    return sampler
+
+
+def generate_initial_points(
+    num_initial_points: int,
+    initial_sampler: InitialPointSampler,
+    space: SearchSpace,
+    target_func: AcquisitionFunction,
+    vectorization: int = 1,
+) -> TensorType:
+    """
+    Return the best starting points for an optimization from those generated by a given sampler.
+
+    :param num_initial_points: Number of best starting points to return.
+    :param initial_sampler: Initial point sampler.
+    :param space: Search space.
+    :param target_func: Target function being optimized.
+    :param vectorization: Vectorization of the target function.
+    """
+    top_fun_values: Optional[TensorType] = None  # [V, num_optimization_runs]
+    top_candidates: Optional[TensorType] = None  # [V, num_optimization_runs, D]
+
+    for candidates in initial_sampler(space):
+        if tf.rank(candidates) == 3:
+            # If samples is a tensor of rank 3, then it is a batch of samples. In this case
+            # the vectorization of the target function must be a multiple of the length of the
+            # second (batch) dimension.
+            remainder = vectorization % tf.shape(candidates)[1]
+            tf.debugging.assert_equal(
+                remainder,
+                tf.constant(0, dtype=remainder.dtype),
+                message=(
+                    f"""
+                    The vectorization of the target function {vectorization} must be a multiple of
+                    the batch shape of initial samples {tf.shape(candidates)[1]}.
+                    """
+                ),
+            )
+            multiple = vectorization // tf.shape(candidates)[1]
+            tiled_candidates = tf.tile(candidates, [1, multiple, 1])  # [samples, V, D]
+        else:
+            tf.debugging.assert_rank(
+                candidates,
+                2,
+                message=(
+                    f"""
+                    The initial samples must be a tensor of rank 2, got a tensor of rank
+                    {tf.rank(candidates)}.
+                    """
+                ),
+            )
+            tiled_candidates = tf.tile(
+                candidates[:, None, :], [1, vectorization, 1]
+            )  # [samples, V, D]
+
+        target_func_values = target_func(tiled_candidates)  # [samples, V]
+        tf.debugging.assert_shapes(
+            [(target_func_values, ("_", vectorization))],
+            message=(
+                f"""
+                The result of function target_func has shape
+                {tf.shape(target_func_values)}, however, expected a trailing
+                dimension of size {vectorization}.
+                """
+            ),
+        )
+
+        # now that we know the output dimension and dtypes, initialize empty top tensors
+        if top_candidates is None:
+            top_candidates = tf.zeros(
+                [vectorization, 0, tf.shape(candidates)[-1]], dtype=candidates.dtype
+            )
+        if top_fun_values is None:
+            top_fun_values = tf.zeros([vectorization, 0], dtype=target_func_values.dtype)
+
+        top_candidates = tf.concat(
+            [top_candidates, tf.transpose(tiled_candidates, [1, 0, 2])], 1
+        )  # [V, samples+num_initial_points, D]
+        top_fun_values = tf.concat(
+            [top_fun_values, tf.transpose(target_func_values)], 1
+        )  # [V, samples+num_initial_points]
+
+        _, top_k_indices = tf.math.top_k(
+            top_fun_values, k=min(num_initial_points, tf.shape(top_fun_values)[-1])
+        )  # [V, num_initial_points]
+
+        top_candidates = tf.gather(
+            top_candidates, top_k_indices, batch_dims=1
+        )  # [V, num_initial_points, D]
+        top_fun_values = tf.gather(
+            top_fun_values, top_k_indices, batch_dims=1
+        )  # [V, num_initial_points]
+
+    if top_candidates is None:
+        raise ValueError("No initial point generated!")
+
+    initial_points = tf.transpose(top_candidates, [1, 0, 2])  # [num_initial_points,V,D]
+    return initial_points
+
+
 def generate_continuous_optimizer(
-    num_initial_samples: int = NUM_SAMPLES_MIN,
+    num_initial_samples: int | InitialPointSampler = NUM_SAMPLES_MIN,
     num_optimization_runs: int = 10,
     num_recovery_runs: int = 10,
     optimizer_args: Optional[dict[str, Any]] = None,
-) -> AcquisitionOptimizer[Box | TaggedProductSearchSpace]:
+) -> AcquisitionOptimizer[Box | CollectionSearchSpace]:
     """
-    Generate a gradient-based optimizer for :class:'Box' and :class:'TaggedProductSearchSpace'
-    spaces and batches of size 1. In the case of a :class:'TaggedProductSearchSpace', We perform
+    Generate a gradient-based optimizer for :class:'Box' and :class:'CollectionSearchSpace'
+    spaces and batches of size 1. In the case of a :class:'CollectionSearchSpace', We perform
     gradient-based optimization across all :class:'Box' subspaces, starting from the best location
     found across a sample of `num_initial_samples` random points.
 
@@ -189,8 +334,14 @@ def generate_continuous_optimizer(
     If all `num_optimization_runs` optimizations fail to converge then we run
     `num_recovery_runs` additional runs starting from random locations (also ran in parallel).
 
-    :param num_initial_samples: The size of the random sample used to find the starting point(s) of
-        the optimization.
+    **Note:** using a large number of `num_initial_samples` and `num_optimization_runs` with a
+    high-dimensional search space can consume a large amount of CPU memory (RAM).
+
+    :param num_initial_samples: The starting point(s) of the optimization. This can be either
+        the number of random samples to use, or a function that given the search space returns
+        the points to use. The latter can be used for example to add pre-optimized starting points
+        to the random points, as well as to batch point generation to reduce memory usage for
+        high-dimensional problems.
     :param num_optimization_runs: The number of separate optimizations to run.
     :param num_recovery_runs: The maximum number of recovery optimization runs in case of failure.
     :param optimizer_args: The keyword arguments to pass to the Scipy L-BFGS-B optimizer.
@@ -198,13 +349,10 @@ def generate_continuous_optimizer(
         can be passed. Note that method, jac and bounds cannot/should not be changed.
     :return: The acquisition optimizer.
     """
-    if num_initial_samples <= 0:
-        raise ValueError(f"num_initial_samples must be positive, got {num_initial_samples}")
-
-    if num_optimization_runs < 0:
+    if num_optimization_runs <= 0:
         raise ValueError(f"num_optimization_runs must be positive, got {num_optimization_runs}")
 
-    if num_initial_samples < num_optimization_runs:
+    if not callable(num_initial_samples) and num_initial_samples < num_optimization_runs:
         raise ValueError(
             f"""
             num_initial_samples {num_initial_samples} must be at
@@ -212,21 +360,27 @@ def generate_continuous_optimizer(
             """
         )
 
-    if num_recovery_runs <= -1:
+    if num_recovery_runs < 0:
         raise ValueError(f"num_recovery_runs must be zero or greater, got {num_recovery_runs}")
 
+    initial_sampler = (
+        sample_from_space(num_initial_samples)
+        if not callable(num_initial_samples)
+        else num_initial_samples
+    )
+
     def optimize_continuous(
-        space: Box | TaggedProductSearchSpace,
+        space: Box | CollectionSearchSpace,
         target_func: Union[AcquisitionFunction, Tuple[AcquisitionFunction, int]],
     ) -> TensorType:
         """
         A gradient-based :const:`AcquisitionOptimizer` for :class:'Box'
-        and :class:`TaggedProductSearchSpace' spaces.
+        and :class:`CollectionSearchSpace' spaces.
 
-        For :class:'TaggedProductSearchSpace' we only apply gradient updates to
+        For :class:'CollectionSearchSpace' we only apply gradient updates to
         its class:'Box' subspaces.
 
-        When this functions receives an acquisition-integer tuple as its `target_func`,it
+        When this function receives an acquisition-integer tuple as its `target_func`,it
         optimizes each of the individual V functions making up `target_func`, i.e.
         evaluating `num_initial_samples` samples, running `num_optimization_runs` runs, and
         (if necessary) running `num_recovery_runs` recovery run for each of the individual
@@ -243,33 +397,18 @@ def generate_continuous_optimizer(
         else:
             V = 1
 
-        if V < 0:
+        if V <= 0:
             raise ValueError(f"vectorization must be positive, got {V}")
 
-        candidates = space.sample(num_initial_samples)[:, None, :]  # [num_initial_samples, 1, D]
-        tiled_candidates = tf.tile(candidates, [1, V, 1])  # [num_initial_samples, V, D]
+        initial_points = generate_initial_points(
+            num_optimization_runs, initial_sampler, space, target_func, V
+        )  # [num_optimization_runs,V,D]
 
-        target_func_values = target_func(tiled_candidates)  # [num_samples, V]
-        tf.debugging.assert_shapes(
-            [(target_func_values, ("_", V))],
-            message=(
-                f"""
-                The result of function target_func has shape
-                {tf.shape(target_func_values)}, however, expected a trailing
-                dimension of size {V}.
-                """
-            ),
-        )
-
-        _, top_k_indices = tf.math.top_k(
-            tf.transpose(target_func_values), k=num_optimization_runs
-        )  # [1, num_optimization_runs] or [V, num_optimization_runs]
-
-        tiled_candidates = tf.transpose(tiled_candidates, [1, 0, 2])  # [V, num_initial_samples, D]
-        top_k_points = tf.gather(
-            tiled_candidates, top_k_indices, batch_dims=1
-        )  # [V, num_optimization_runs, D]
-        initial_points = tf.transpose(top_k_points, [1, 0, 2])  # [num_optimization_runs,V,D]
+        if len(initial_points) < num_optimization_runs:
+            raise ValueError(
+                f"Not enough initial points generated ({len(initial_points)} "
+                f"for {num_optimization_runs} optimization runs)"
+            )
 
         (
             successes,
@@ -289,11 +428,42 @@ def generate_continuous_optimizer(
         total_nfev = tf.reduce_max(nfev)  # acquisition function is evaluated in parallel
 
         recovery_run = False
-        if (
-            num_recovery_runs and not successful_optimization
-        ):  # if all optimizations failed for a function then try again from random starts
-            random_points = space.sample(num_recovery_runs)[:, None, :]  # [num_recovery_runs, 1, D]
-            tiled_random_points = tf.tile(random_points, [1, V, 1])  # [num_recovery_runs, V, D]
+        if num_recovery_runs and not successful_optimization:
+            # if all optimizations failed for a function then try again from random starts
+            random_points = space.sample(num_recovery_runs)
+            if tf.rank(random_points) == 3:
+                # If samples is a tensor of rank 3, then it is a batch of samples. In this case
+                # the vectorization of the target function must be a multiple of the length of the
+                # second (batch) dimension.
+                remainder = V % tf.shape(random_points)[1]
+                tf.debugging.assert_equal(
+                    remainder,
+                    tf.constant(0, dtype=remainder.dtype),
+                    message=(
+                        f"""
+                        The vectorization of the target function {V} must be a multiple of the batch
+                        shape of random samples {tf.shape(random_points)[1]}.
+                        """
+                    ),
+                )
+                multiple = V // tf.shape(random_points)[1]
+                tiled_random_points = tf.tile(
+                    random_points, [1, multiple, 1]  # [num_recovery_runs, V, D]
+                )
+            else:
+                tf.debugging.assert_rank(
+                    random_points,
+                    2,
+                    message=(
+                        f"""
+                        The random samples must be a tensor of rank 2, got a tensor of rank
+                        {tf.rank(random_points)}.
+                        """
+                    ),
+                )
+                tiled_random_points = tf.tile(
+                    random_points[:, None, :], [1, V, 1]
+                )  # [num_recovery_runs, V, D]
 
             (
                 recovery_successes,
@@ -350,7 +520,7 @@ def generate_continuous_optimizer(
                 if V == 1:
                     logging.scalar("spo_improvement_on_initial_samples", improvements)
                 else:
-                    logging.histogram("spo_improvement_on_initial_samples", improvements)
+                    logging.histogram("spo_improvements_on_initial_samples", improvements)
 
         best_run_ids = tf.math.argmax(fun_values, axis=0)  # [V]
         chosen_points = tf.gather(
@@ -383,7 +553,7 @@ def _perform_parallel_continuous_optimization(
     of our acquisition function (and its gradients) is calculated in parallel
     (for each optimization step) using Tensorflow.
 
-    For :class:'TaggedProductSearchSpace' we only apply gradient updates to
+    For :class:'CollectionSearchSpace' we only apply gradient updates to
     its :class:'Box' subspaces, fixing the discrete elements to the best values
     found across the initial random search. To fix these discrete elements, we
     optimize over a continuous :class:'Box' relaxation of the discrete subspaces
@@ -427,15 +597,8 @@ def _perform_parallel_continuous_optimization(
     def _objective_value_and_gradient(x: TensorType) -> Tuple[TensorType, TensorType]:
         return tfp.math.value_and_gradient(_objective_value, x)  # [len(x), 1], [len(x), D]
 
-    if isinstance(
-        space, TaggedProductSearchSpace
-    ):  # build continuous relaxation of discrete subspaces
-        bounds = [
-            get_bounds_of_box_relaxation_around_point(space, vectorized_starting_points[i : i + 1])
-            for i in tf.range(num_optimization_runs)
-        ]
-    else:
-        bounds = [spo.Bounds(space.lower, space.upper)] * num_optimization_runs
+    # Get the bounds for the optimization
+    bounds = get_bounds_of_optimization(space, starting_points)
 
     # Initialize the numpy arrays to be passed to the greenlets
     np_batch_x = np.zeros((num_optimization_runs, tf.shape(starting_points)[-1]), dtype=np.float64)
@@ -518,6 +681,7 @@ class ScipyOptimizerGreenlet(gr.greenlet):  # type: ignore[misc]
         constraints: Sequence[Constraint],
         optimizer_args: Optional[dict[str, Any]] = None,
     ) -> spo.OptimizeResult:
+        """Greenlet run method."""
         cache_x = start + 1  # Any value different from `start`.
         cache_y: Optional["np.ndarray[Any, Any]"] = None
         cache_dy_dx: Optional["np.ndarray[Any, Any]"] = None
@@ -539,7 +703,7 @@ class ScipyOptimizerGreenlet(gr.greenlet):  # type: ignore[misc]
 
         method = "trust-constr" if len(constraints) else "l-bfgs-b"
         optimizer_args = dict(
-            dict(method=method, constraints=constraints), **(optimizer_args or {})
+            {"method": method, "constraints": constraints}, **(optimizer_args or {})
         )
         return spo.minimize(
             lambda x: value_and_gradient(x)[0],
@@ -577,6 +741,126 @@ def get_bounds_of_box_relaxation_around_point(
     return spo.Bounds(space_with_fixed_discrete.lower, space_with_fixed_discrete.upper)
 
 
+@check_shapes(
+    "starting_points: [num_optimization_runs, V, D]",
+)
+def get_bounds_of_optimization(space: SearchSpace, starting_points: TensorType) -> List[spo.Bounds]:
+    """
+    Returns a list of bounds for all the optimization runs, to be used in the Scipy optimizer.
+    The bounds are based on the provided search space and the starting points. The length of the
+    list is equal to the total number of individual optimization runs,
+    i.e. the number of starting points: `num_optimization_runs` x `V`.
+
+    For each starting point, the bounds are obtained as follows depending on the type of search
+    space:
+    - For a `TaggedProductSearchSpace`, a "continuous relaxation" of the discrete subspaces is
+    built by creating bounds around the point. The discrete components of the created
+    search space are fixed at the respective component of the point, and the
+    remaining continuous components are set to the original bounds of the search space.
+    - For a `TaggedMultiSearchSpace`, the bounds for each point are obtained in a similar
+    manner as above, but based potentially on a different subspace for each point;
+    instead of sharing a single set of bounds from the common search space.
+    The subspaces are assigned to the optimization runs in a round-robin fashion along
+    dimension 1 (of size `V`) of the starting points. An error is raised if size `V` is not a
+    multiple of the number of subspaces.
+    - For other types of search spaces, the original bounds are used for each optimization run.
+
+    :param space: The original search space.
+    :param starting_points: The points at which to begin the optimizations, with shape
+        `[num_optimization_runs, V, D]`. The leading dimension of `starting_points` controls the
+        number of individual optimization runs for each of the `V` starting points or target
+        functions.
+    :return: A list of bounds for the Scipy optimizer. The length of the list
+        is equal to the number of individual optimization runs, i.e. `num_optimization_runs` x `V`.
+
+    For example, for a 2D `TaggedMultiSearchSpace` with two subspaces, each with a continuous
+    and a discrete component::
+
+        space = TaggedMultiSearchSpace(
+            [
+                TaggedProductSearchSpace(
+                    [Box([0], [1]), DiscreteSearchSpace([[11], [15], [21], [25], [31], [35]])]
+                ),
+                TaggedProductSearchSpace(
+                    [Box([2], [3]), DiscreteSearchSpace([[13], [17], [23], [27], [33], [37]])]
+                )
+            ]
+        )
+
+    Consider 2 optimization runs per point and a vectorization `V` of 4, for a total of 8
+    optimization runs. Given the following starting points::
+
+        starting_points = tf.constant(
+            [
+                [[10, 11], [12, 13], [14, 15], [16, 17]],
+                [[20, 21], [22, 23], [24, 25], [26, 27]],
+            ],
+            dtype=tf.float64
+        )
+
+    The returned list of bounds for the optimization would be::
+
+        [
+            spo.Bounds([0, 11], [1, 11]),  # For point at index [0, 0], using subspace 0
+            spo.Bounds([2, 13], [3, 13]),  #         //         [0, 1],       //       1
+            spo.Bounds([0, 15], [1, 15]),  #         //         [0, 2],       //       0
+            spo.Bounds([2, 17], [3, 17]),  #         //         [0, 3],       //       1
+            spo.Bounds([0, 21], [1, 21]),  #         //         [1, 0],       //       0
+            spo.Bounds([2, 23], [3, 23]),  #         //         [1, 1],       //       1
+            spo.Bounds([0, 25], [1, 25]),  #         //         [1, 2],       //       0
+            spo.Bounds([2, 27], [3, 27]),  #         //         [1, 3],       //       1
+        ]
+    """
+
+    num_optimization_runs_per_function, V, D = starting_points.shape
+    num_total_optimization_runs = num_optimization_runs_per_function * V
+
+    if isinstance(
+        space, TaggedProductSearchSpace
+    ):  # build continuous relaxation of discrete subspaces
+        vectorized_starting_points = tf.reshape(
+            starting_points, [-1, D]
+        )  # [num_total_optimization_runs, D]
+        bounds = [
+            get_bounds_of_box_relaxation_around_point(space, vectorized_starting_points[i : i + 1])
+            for i in tf.range(num_total_optimization_runs)
+        ]
+    elif isinstance(space, TaggedMultiSearchSpace):
+        subspaces = [space.get_subspace(tag) for tag in space.subspace_tags]
+        # The vectorization `V` of the target function must be a multple of the number of subspaces.
+        remainder = V % len(subspaces)
+        tf.debugging.assert_equal(
+            remainder,
+            0,
+            message=(
+                f"""
+                The vectorization of the target function {V} must be a multiple of the number of
+                subspaces {len(subspaces)}.
+                """
+            ),
+        )
+        multiple = V // len(subspaces)
+        subspaces = subspaces * multiple
+
+        # Create a sequence of bounds for each optimization run, per subspace.
+        # For product subspaces, we build a continuous relaxation of the discrete subspaces.
+        # Otherwise, we use the original bounds.
+        bounds = [
+            [
+                get_bounds_of_box_relaxation_around_point(ss, starting_points[i : i + 1, j])
+                if isinstance(ss, TaggedProductSearchSpace)
+                else spo.Bounds(ss.lower, ss.upper)
+                for j, ss in enumerate(subspaces)
+            ]
+            for i in tf.range(num_optimization_runs_per_function)
+        ]
+        bounds = list(chain.from_iterable(bounds))
+    else:
+        bounds = [spo.Bounds(space.lower, space.upper)] * num_total_optimization_runs
+
+    return bounds
+
+
 def batchify_joint(
     batch_size_one_optimizer: AcquisitionOptimizer[SearchSpaceType],
     batch_size: int,
