@@ -26,7 +26,7 @@ import tensorflow_probability as tfp
 from scipy.optimize import bisect
 
 from ..models import ProbabilisticModel
-from ..models.interfaces import HasTrajectorySampler, ProbabilisticModelType
+from ..models.interfaces import HasTrajectorySampler, ProbabilisticModelType, SupportsPredictY
 from ..types import TensorType
 from .utils import select_nth_output
 
@@ -46,6 +46,8 @@ class ThompsonSampler(ABC, Generic[ProbabilisticModelType]):
 
     @property
     def sample_min_value(self) -> bool:
+        """Whether this samples from the minimum value of the function
+        (as opposed to the function's minimiser)."""
         return self._sample_min_value
 
     def __repr__(self) -> str:
@@ -92,7 +94,7 @@ class ExactThompsonSampler(ThompsonSampler[ProbabilisticModel]):
     ) -> TensorType:
         """
         Return exact samples from either the objective function's minimiser or its minimal value
-        over the candidate set `at`.
+        over the candidate set `at`. Note that minimiser ties aren't broken randomly.
 
         :param model: The model to sample from.
         :param sample_size: The desired number of samples.
@@ -174,15 +176,17 @@ class GumbelSampler(ThompsonSampler[ProbabilisticModel]):
         tf.debugging.assert_positive(sample_size)
         tf.debugging.assert_shapes([(at, ["N", None])])
 
-        try:
+        if isinstance(model, SupportsPredictY):
             fmean, fvar = model.predict_y(at)
-        except NotImplementedError:
+        else:
             fmean, fvar = model.predict(at)
 
         fsd = tf.math.sqrt(fvar)
 
         def probf(y: tf.Tensor) -> tf.Tensor:  # Build empirical CDF for Pr(y*^hat<y)
-            unit_normal = tfp.distributions.Normal(tf.cast(0, fmean.dtype), tf.cast(1, fmean.dtype))
+            unit_normal = tfp.distributions.Normal(
+                tf.constant(0, fmean.dtype), tf.constant(1, fmean.dtype)
+            )
             log_cdf = unit_normal.log_cdf(-(y - fmean) / fsd)
             return 1 - tf.exp(tf.reduce_sum(log_cdf, axis=0))
 
@@ -225,7 +229,7 @@ class ThompsonSamplerFromTrajectory(ThompsonSampler[HasTrajectorySampler]):
     ) -> TensorType:
         """
         Return approximate samples from either the objective function's minimser or its minimal
-        value over the candidate set `at`.
+        value over the candidate set `at`. Note that minimiser ties aren't broken randomly.
 
         :param model: The model to sample from.
         :param sample_size: The desired number of samples.
@@ -245,7 +249,7 @@ class ThompsonSamplerFromTrajectory(ThompsonSampler[HasTrajectorySampler]):
         if not isinstance(model, HasTrajectorySampler):
             raise ValueError(
                 f"Thompson sampling from trajectory only supports models with a trajectory_sampler "
-                f"method; received {model.__repr__()}"
+                f"method; received {model!r}"
             )
 
         trajectory_sampler = model.trajectory_sampler()
