@@ -24,6 +24,7 @@ from collections import Counter
 from dataclasses import dataclass
 from pathlib import Path
 from typing import (
+    Any,
     Callable,
     ClassVar,
     Dict,
@@ -43,6 +44,7 @@ import tensorflow as tf
 from scipy.spatial.distance import pdist
 
 from .acquisition.multi_objective import non_dominated
+from .models.utils import optimize_model_and_save_result
 
 try:
     import pandas as pd
@@ -52,13 +54,24 @@ except ModuleNotFoundError:
     sns = None
 
 from . import logging
-from .acquisition.rule import AcquisitionRule, EfficientGlobalOptimization
+from .acquisition.rule import (
+    AcquisitionRule,
+    EfficientGlobalOptimization,
+    LocalDatasetsAcquisitionRule,
+)
+from .acquisition.utils import with_local_datasets
 from .data import Dataset
-from .models import SupportsCovarianceWithTopFidelity, TrainableProbabilisticModel
+from .models import (
+    ProbabilisticModel,
+    SupportsCovarianceWithTopFidelity,
+    TrainableProbabilisticModel,
+)
+from .objectives.utils import mk_batch_observer
 from .observer import OBJECTIVE, Observer
 from .space import SearchSpace
 from .types import State, Tag, TensorType
 from .utils import Err, Ok, Result, Timer
+from .utils.misc import LocalizedTag, get_value_for_tag, ignoring_local_tags
 
 StateType = TypeVar("StateType")
 """ Unbound type variable. """
@@ -66,6 +79,13 @@ StateType = TypeVar("StateType")
 SearchSpaceType = TypeVar("SearchSpaceType", bound=SearchSpace)
 """ Type variable bound to :class:`SearchSpace`. """
 
+ProbabilisticModelType = TypeVar(
+    "ProbabilisticModelType",
+    bound=ProbabilisticModel,
+    covariant=True,
+)
+""" Covariant type variable bound to :class:`ProbabilisticModel`. """
+
 TrainableProbabilisticModelType = TypeVar(
     "TrainableProbabilisticModelType", bound=TrainableProbabilisticModel, contravariant=True
 )
@@ -79,13 +99,13 @@ EarlyStopCallback = Callable[
 
 
 @dataclass(frozen=True)
-class Record(Generic[StateType]):
+class Record(Generic[StateType, ProbabilisticModelType]):
     """Container to record the state of each step of the optimization process."""
 
     datasets: Mapping[Tag, Dataset]
     """ The known data from the observer. """
 
-    models: Mapping[Tag, TrainableProbabilisticModel]
+    models: Mapping[Tag, ProbabilisticModelType]
     """ The models over the :attr:`datasets`. """
 
     acquisition_state: StateType | None
@@ -94,20 +114,24 @@ class Record(Generic[StateType]):
     @property
     def dataset(self) -> Dataset:
         """The dataset when there is just one dataset."""
-        if len(self.datasets) == 1:
-            return next(iter(self.datasets.values()))
+        # Ignore local datasets.
+        datasets: Mapping[Tag, Dataset] = ignoring_local_tags(self.datasets)
+        if len(datasets) == 1:
+            return next(iter(datasets.values()))
         else:
-            raise ValueError(f"Expected a single dataset, found {len(self.datasets)}")
+            raise ValueError(f"Expected a single dataset, found {len(datasets)}")
 
     @property
-    def model(self) -> TrainableProbabilisticModel:
+    def model(self) -> ProbabilisticModelType:
         """The model when there is just one dataset."""
-        if len(self.models) == 1:
-            return next(iter(self.models.values()))
+        # Ignore local models.
+        models: Mapping[Tag, ProbabilisticModelType] = ignoring_local_tags(self.models)
+        if len(models) == 1:
+            return next(iter(models.values()))
         else:
-            raise ValueError(f"Expected a single model, found {len(self.models)}")
+            raise ValueError(f"Expected a single model, found {len(models)}")
 
-    def save(self, path: Path | str) -> FrozenRecord[StateType]:
+    def save(self, path: Path | str) -> FrozenRecord[StateType, ProbabilisticModelType]:
         """Save the record to disk. Will overwrite any existing file at the same path."""
         Path(path).parent.mkdir(exist_ok=True, parents=True)
         with open(path, "wb") as f:
@@ -116,7 +140,7 @@ class Record(Generic[StateType]):
 
 
 @dataclass(frozen=True)
-class FrozenRecord(Generic[StateType]):
+class FrozenRecord(Generic[StateType, ProbabilisticModelType]):
     """
     A Record container saved on disk.
 
@@ -127,7 +151,7 @@ class FrozenRecord(Generic[StateType]):
     path: Path
     """ The path to the pickled Record. """
 
-    def load(self) -> Record[StateType]:
+    def load(self) -> Record[StateType, ProbabilisticModelType]:
         """Load the record into memory."""
         with open(self.path, "rb") as f:
             return dill.load(f)
@@ -138,7 +162,7 @@ class FrozenRecord(Generic[StateType]):
         return self.load().datasets
 
     @property
-    def models(self) -> Mapping[Tag, TrainableProbabilisticModel]:
+    def models(self) -> Mapping[Tag, ProbabilisticModelType]:
         """The models over the :attr:`datasets`."""
         return self.load().models
 
@@ -153,7 +177,7 @@ class FrozenRecord(Generic[StateType]):
         return self.load().dataset
 
     @property
-    def model(self) -> TrainableProbabilisticModel:
+    def model(self) -> ProbabilisticModelType:
         """The model when there is just one dataset."""
         return self.load().model
 
@@ -161,16 +185,18 @@ class FrozenRecord(Generic[StateType]):
 # this should be a generic NamedTuple, but mypy doesn't support them
 #  https://github.com/python/mypy/issues/685
 @dataclass(frozen=True)
-class OptimizationResult(Generic[StateType]):
+class OptimizationResult(Generic[StateType, ProbabilisticModelType]):
     """The final result, and the historical data of the optimization process."""
 
-    final_result: Result[Record[StateType]]
+    final_result: Result[Record[StateType, ProbabilisticModelType]]
     """
     The final result of the optimization process. This contains either a :class:`Record` or an
     exception.
     """
 
-    history: list[Record[StateType] | FrozenRecord[StateType]]
+    history: list[
+        Record[StateType, ProbabilisticModelType] | FrozenRecord[StateType, ProbabilisticModelType]
+    ]
     r"""
     The history of the :class:`Record`\ s from each step of the optimization process. These
     :class:`Record`\ s are created at the *start* of each loop, and as such will never
@@ -187,7 +213,13 @@ class OptimizationResult(Generic[StateType]):
 
     def astuple(
         self,
-    ) -> tuple[Result[Record[StateType]], list[Record[StateType] | FrozenRecord[StateType]]]:
+    ) -> tuple[
+        Result[Record[StateType, ProbabilisticModelType]],
+        list[
+            Record[StateType, ProbabilisticModelType]
+            | FrozenRecord[StateType, ProbabilisticModelType]
+        ],
+    ]:
         """
         **Note:** In contrast to the standard library function :func:`dataclasses.astuple`, this
         method does *not* deepcopy instance attributes.
@@ -196,6 +228,16 @@ class OptimizationResult(Generic[StateType]):
         """
         return self.final_result, self.history
 
+    @property
+    def is_ok(self) -> bool:
+        """`True` if the final result contains a :class:`Record`."""
+        return self.final_result.is_ok
+
+    @property
+    def is_err(self) -> bool:
+        """`True` if the final result contains an exception."""
+        return self.final_result.is_err
+
     def try_get_final_datasets(self) -> Mapping[Tag, Dataset]:
         """
         Convenience method to attempt to get the final data.
@@ -214,6 +256,8 @@ class OptimizationResult(Generic[StateType]):
         :raise ValueError: If the optimization was not a single dataset run.
         """
         datasets = self.try_get_final_datasets()
+        # Ignore local datasets.
+        datasets = ignoring_local_tags(datasets)
         if len(datasets) == 1:
             return next(iter(datasets.values()))
         else:
@@ -239,7 +283,7 @@ class OptimizationResult(Generic[StateType]):
         arg_min_idx = tf.squeeze(tf.argmin(dataset.observations, axis=0))
         return dataset.query_points[arg_min_idx], dataset.observations[arg_min_idx], arg_min_idx
 
-    def try_get_final_models(self) -> Mapping[Tag, TrainableProbabilisticModel]:
+    def try_get_final_models(self) -> Mapping[Tag, ProbabilisticModelType]:
         """
         Convenience method to attempt to get the final models.
 
@@ -248,7 +292,7 @@ class OptimizationResult(Generic[StateType]):
         """
         return self.final_result.unwrap().models
 
-    def try_get_final_model(self) -> TrainableProbabilisticModel:
+    def try_get_final_model(self) -> ProbabilisticModelType:
         """
         Convenience method to attempt to get the final model for a single model run.
 
@@ -257,13 +301,15 @@ class OptimizationResult(Generic[StateType]):
         :raise ValueError: If the optimization was not a single model run.
         """
         models = self.try_get_final_models()
+        # Ignore local models.
+        models = ignoring_local_tags(models)
         if len(models) == 1:
             return next(iter(models.values()))
         else:
             raise ValueError(f"Expected single model, found {len(models)}")
 
     @property
-    def loaded_history(self) -> list[Record[StateType]]:
+    def loaded_history(self) -> list[Record[StateType, ProbabilisticModelType]]:
         """The history of the optimization process loaded into memory."""
         return [record if isinstance(record, Record) else record.load() for record in self.history]
 
@@ -283,7 +329,9 @@ class OptimizationResult(Generic[StateType]):
             record.save(record_path)
 
     @classmethod
-    def from_path(cls, base_path: Path | str) -> OptimizationResult[StateType]:
+    def from_path(
+        cls, base_path: Path | str
+    ) -> OptimizationResult[StateType, ProbabilisticModelType]:
         """Load a previously saved OptimizationResult."""
         try:
             with open(Path(base_path) / cls.RESULTS_FILENAME, "rb") as f:
@@ -291,9 +339,10 @@ class OptimizationResult(Generic[StateType]):
         except FileNotFoundError as e:
             result = Err(e)
 
-        history: list[Record[StateType] | FrozenRecord[StateType]] = [
-            FrozenRecord(file) for file in sorted(Path(base_path).glob(cls.STEP_GLOB))
-        ]
+        history: list[
+            Record[StateType, ProbabilisticModelType]
+            | FrozenRecord[StateType, ProbabilisticModelType]
+        ] = [FrozenRecord(file) for file in sorted(Path(base_path).glob(cls.STEP_GLOB))]
         return cls(result, history)
 
 
@@ -326,11 +375,13 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
         *,
         track_state: bool = True,
         track_path: Optional[Path | str] = None,
+        fit_model: bool = True,
         fit_initial_model: bool = True,
         early_stop_callback: Optional[
             EarlyStopCallback[TrainableProbabilisticModel, object]
         ] = None,
-    ) -> OptimizationResult[None]:
+        start_step: int = 0,
+    ) -> OptimizationResult[None, TrainableProbabilisticModel]:
         ...
 
     @overload
@@ -345,14 +396,16 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
         *,
         track_state: bool = True,
         track_path: Optional[Path | str] = None,
+        fit_model: bool = True,
         fit_initial_model: bool = True,
         early_stop_callback: Optional[
             EarlyStopCallback[TrainableProbabilisticModelType, object]
         ] = None,
+        start_step: int = 0,
         # this should really be OptimizationResult[None], but tf.Tensor is untyped so the type
         # checker can't differentiate between TensorType and State[S | None, TensorType], and
         # the return types clash. object is close enough to None that object will do.
-    ) -> OptimizationResult[object]:
+    ) -> OptimizationResult[object, TrainableProbabilisticModelType]:
         ...
 
     @overload
@@ -367,11 +420,13 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
         *,
         track_state: bool = True,
         track_path: Optional[Path | str] = None,
+        fit_model: bool = True,
         fit_initial_model: bool = True,
         early_stop_callback: Optional[
             EarlyStopCallback[TrainableProbabilisticModelType, object]
         ] = None,
-    ) -> OptimizationResult[object]:
+        start_step: int = 0,
+    ) -> OptimizationResult[object, TrainableProbabilisticModelType]:
         ...
 
     @overload
@@ -387,11 +442,13 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
         *,
         track_state: bool = True,
         track_path: Optional[Path | str] = None,
+        fit_model: bool = True,
         fit_initial_model: bool = True,
         early_stop_callback: Optional[
             EarlyStopCallback[TrainableProbabilisticModelType, StateType]
         ] = None,
-    ) -> OptimizationResult[StateType]:
+        start_step: int = 0,
+    ) -> OptimizationResult[StateType, TrainableProbabilisticModelType]:
         ...
 
     @overload
@@ -407,11 +464,13 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
         *,
         track_state: bool = True,
         track_path: Optional[Path | str] = None,
+        fit_model: bool = True,
         fit_initial_model: bool = True,
         early_stop_callback: Optional[
             EarlyStopCallback[TrainableProbabilisticModelType, StateType]
         ] = None,
-    ) -> OptimizationResult[StateType]:
+        start_step: int = 0,
+    ) -> OptimizationResult[StateType, TrainableProbabilisticModelType]:
         ...
 
     @overload
@@ -423,11 +482,13 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
         *,
         track_state: bool = True,
         track_path: Optional[Path | str] = None,
+        fit_model: bool = True,
         fit_initial_model: bool = True,
         early_stop_callback: Optional[
             EarlyStopCallback[TrainableProbabilisticModel, object]
         ] = None,
-    ) -> OptimizationResult[None]:
+        start_step: int = 0,
+    ) -> OptimizationResult[None, TrainableProbabilisticModel]:
         ...
 
     @overload
@@ -442,11 +503,13 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
         *,
         track_state: bool = True,
         track_path: Optional[Path | str] = None,
+        fit_model: bool = True,
         fit_initial_model: bool = True,
         early_stop_callback: Optional[
             EarlyStopCallback[TrainableProbabilisticModelType, object]
         ] = None,
-    ) -> OptimizationResult[object]:
+        start_step: int = 0,
+    ) -> OptimizationResult[object, TrainableProbabilisticModelType]:
         ...
 
     @overload
@@ -461,11 +524,13 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
         *,
         track_state: bool = True,
         track_path: Optional[Path | str] = None,
+        fit_model: bool = True,
         fit_initial_model: bool = True,
         early_stop_callback: Optional[
             EarlyStopCallback[TrainableProbabilisticModelType, object]
         ] = None,
-    ) -> OptimizationResult[object]:
+        start_step: int = 0,
+    ) -> OptimizationResult[object, TrainableProbabilisticModelType]:
         ...
 
     @overload
@@ -481,11 +546,13 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
         *,
         track_state: bool = True,
         track_path: Optional[Path | str] = None,
+        fit_model: bool = True,
         fit_initial_model: bool = True,
         early_stop_callback: Optional[
             EarlyStopCallback[TrainableProbabilisticModelType, StateType]
         ] = None,
-    ) -> OptimizationResult[StateType]:
+        start_step: int = 0,
+    ) -> OptimizationResult[StateType, TrainableProbabilisticModelType]:
         ...
 
     @overload
@@ -501,11 +568,13 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
         *,
         track_state: bool = True,
         track_path: Optional[Path | str] = None,
+        fit_model: bool = True,
         fit_initial_model: bool = True,
         early_stop_callback: Optional[
             EarlyStopCallback[TrainableProbabilisticModelType, StateType]
         ] = None,
-    ) -> OptimizationResult[StateType]:
+        start_step: int = 0,
+    ) -> OptimizationResult[StateType, TrainableProbabilisticModelType]:
         ...
 
     def optimize(
@@ -523,11 +592,16 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
         *,
         track_state: bool = True,
         track_path: Optional[Path | str] = None,
+        fit_model: bool = True,
         fit_initial_model: bool = True,
         early_stop_callback: Optional[
             EarlyStopCallback[TrainableProbabilisticModelType, StateType]
         ] = None,
-    ) -> OptimizationResult[StateType] | OptimizationResult[None]:
+        start_step: int = 0,
+    ) -> (
+        OptimizationResult[StateType, TrainableProbabilisticModelType]
+        | OptimizationResult[None, TrainableProbabilisticModelType]
+    ):
         """
         Attempt to find the minimizer of the ``observer`` in the ``search_space`` (both specified at
         :meth:`__init__`). This is the central implementation of the Bayesian optimization loop.
@@ -568,12 +642,16 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
             step. Models and acquisition state are copied using `copy.deepcopy`.
         :param track_path: If set, the optimization state is saved to disk at this path,
             rather than being copied in memory.
-        :param fit_initial_model: If `False`, this method assumes that the initial models have
-            already been optimized on the datasets and so do not require optimization before the
-            first optimization step.
+        :param fit_model: If `False` then we never fit the model during BO (e.g. if we
+            are using a rule that doesn't rely on the models and don't want to waste computation).
+        :param fit_initial_model: If `False` then we assume that the initial models have
+            already been optimized on the datasets and so do not require optimization before
+            the first optimization step.
         :param early_stop_callback: An optional callback that is evaluated with the current
             datasets, models and optimization state before every optimization step. If this
             returns `True` then the optimization loop is terminated early.
+        :param start_step: The step number to start with. This number is removed from ``num_steps``
+            and is useful for restarting previous computations.
         :return: An :class:`OptimizationResult`. The :attr:`final_result` element contains either
             the final optimization data, models and acquisition state, or, if an exception was
             raised while executing the optimization loop, it contains the exception raised. In
@@ -587,10 +665,15 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
             - ``datasets`` or ``models`` are empty
             - the default `acquisition_rule` is used and the tags are not `OBJECTIVE`.
         """
+        # Copy the dataset so we don't change the one provided by the user.
+        datasets = copy.deepcopy(datasets)
+
         if isinstance(datasets, Dataset):
             datasets = {OBJECTIVE: datasets}
-            models = {OBJECTIVE: models}  # type: ignore[dict-item]
+        if not isinstance(models, Mapping):
+            models = {OBJECTIVE: models}
 
+        filtered_datasets = datasets
         # reassure the type checker that everything is tagged
         datasets = cast(Dict[Tag, Dataset], datasets)
         models = cast(Dict[Tag, TrainableProbabilisticModelType], models)
@@ -598,10 +681,14 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
         if num_steps < 0:
             raise ValueError(f"num_steps must be at least 0, got {num_steps}")
 
-        if datasets.keys() != models.keys():
+        # Get set of dataset and model keys, ignoring any local tag index. That is, only the
+        # global tag part is considered.
+        datasets_keys = {LocalizedTag.from_tag(tag).global_tag for tag in datasets.keys()}
+        models_keys = {LocalizedTag.from_tag(tag).global_tag for tag in models.keys()}
+        if datasets_keys != models_keys:
             raise ValueError(
-                f"datasets and models should contain the same keys. Got {datasets.keys()} and"
-                f" {models.keys()} respectively."
+                f"datasets and models should contain the same keys. Got {datasets_keys} and"
+                f" {models_keys} respectively."
             )
 
         if not datasets:
@@ -618,7 +705,11 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
                 SearchSpaceType, TrainableProbabilisticModelType
             ]()
 
-        history: list[FrozenRecord[StateType] | Record[StateType]] = []
+        history: list[
+            FrozenRecord[StateType, TrainableProbabilisticModelType]
+            | Record[StateType, TrainableProbabilisticModelType]
+        ] = []
+        history_state = acquisition_state
         query_plot_dfs: dict[int, pd.DataFrame] = {}
         observation_plot_dfs = observation_plot_init(datasets)
 
@@ -634,7 +725,7 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
                     num_steps,
                 )
 
-        for step in range(1, num_steps + 1):
+        for step in range(start_step + 1, num_steps + 1):
             logging.set_step_number(step)
 
             if early_stop_callback and early_stop_callback(datasets, models, acquisition_state):
@@ -644,15 +735,18 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
             try:
                 if track_state:
                     try:
+                        # note that we use the state as it was at acquisition time, not the one
+                        # modified in filter_datasets in preparation for the next acquisition,
+                        # so we can restart the optimization correctly (and also for plotting)
                         if track_path is None:
                             datasets_copy = copy.deepcopy(datasets)
                             models_copy = copy.deepcopy(models)
-                            acquisition_state_copy = copy.deepcopy(acquisition_state)
-                            record = Record(datasets_copy, models_copy, acquisition_state_copy)
+                            history_state_copy = copy.deepcopy(history_state)
+                            record = Record(datasets_copy, models_copy, history_state_copy)
                             history.append(record)
                         else:
                             track_path = Path(track_path)
-                            record = Record(datasets, models, acquisition_state)
+                            record = Record(datasets, models, history_state)
                             file_name = OptimizationResult.step_filename(step, num_steps)
                             history.append(record.save(track_path / file_name))
                     except Exception as e:
@@ -667,31 +761,57 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
                             "will be available."
                         ) from e
 
-                if step == 1 and fit_initial_model:
-                    with Timer() as initial_model_fitting_timer:
-                        for tag, model in models.items():
-                            dataset = datasets[tag]
-                            model.update(dataset)
-                            model.optimize(dataset)
-                    if summary_writer:
-                        logging.set_step_number(0)
-                        with summary_writer.as_default(step=0):
-                            write_summary_initial_model_fit(
-                                datasets, models, initial_model_fitting_timer
-                            )
-                        logging.set_step_number(step)
+                if step == 1:
+                    # See explanation in AskTellOptimizer.__init__().
+                    if isinstance(acquisition_rule, LocalDatasetsAcquisitionRule):
+                        datasets = with_local_datasets(
+                            datasets, acquisition_rule.num_local_datasets
+                        )
+                        acquisition_rule.initialize_subspaces(self._search_space)
+
+                    filtered_datasets_or_callable: Mapping[Tag, Dataset] | State[
+                        StateType | None, Mapping[Tag, Dataset]
+                    ] = acquisition_rule.filter_datasets(models, datasets)
+                    if callable(filtered_datasets_or_callable):
+                        acquisition_state, filtered_datasets = filtered_datasets_or_callable(
+                            acquisition_state
+                        )
+                    else:
+                        filtered_datasets = filtered_datasets_or_callable
+
+                    if fit_model and fit_initial_model:
+                        with Timer() as initial_model_fitting_timer:
+                            for tag, model in models.items():
+                                # Prefer local dataset if available.
+                                tags = [tag, LocalizedTag.from_tag(tag).global_tag]
+                                _, dataset = get_value_for_tag(filtered_datasets, *tags)
+                                assert dataset is not None
+                                model.update(dataset)
+                                optimize_model_and_save_result(model, dataset)
+                        if summary_writer:
+                            logging.set_step_number(0)
+                            with summary_writer.as_default(step=0):
+                                write_summary_initial_model_fit(
+                                    datasets, models, initial_model_fitting_timer
+                                )
+                            logging.set_step_number(step)
 
                 with Timer() as total_step_wallclock_timer:
                     with Timer() as query_point_generation_timer:
                         points_or_stateful = acquisition_rule.acquire(
-                            self._search_space, models, datasets=datasets
+                            self._search_space, models, datasets=filtered_datasets
                         )
                         if callable(points_or_stateful):
                             acquisition_state, query_points = points_or_stateful(acquisition_state)
+                            history_state = acquisition_state
                         else:
                             query_points = points_or_stateful
 
-                    observer_output = self._observer(query_points)
+                    observer = self._observer
+                    # If query_points are rank 3, then use a batched observer.
+                    if tf.rank(query_points) == 3:
+                        observer = mk_batch_observer(observer)
+                    observer_output = observer(query_points)
 
                     tagged_output = (
                         observer_output
@@ -699,13 +819,27 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
                         else {OBJECTIVE: observer_output}
                     )
 
-                    datasets = {tag: datasets[tag] + tagged_output[tag] for tag in tagged_output}
+                    for tag, new_dataset in tagged_output.items():
+                        datasets[tag] += new_dataset
+
+                    filtered_datasets_or_callable = acquisition_rule.filter_datasets(
+                        models, datasets
+                    )
+                    if callable(filtered_datasets_or_callable):
+                        acquisition_state, filtered_datasets = filtered_datasets_or_callable(
+                            acquisition_state
+                        )
+                    else:
+                        filtered_datasets = filtered_datasets_or_callable
 
                     with Timer() as model_fitting_timer:
-                        for tag, model in models.items():
-                            dataset = datasets[tag]
-                            model.update(dataset)
-                            model.optimize(dataset)
+                        if fit_model:
+                            for tag, model in models.items():
+                                # Always use the matching dataset to the model. If the model is
+                                # local, then the dataset should be too by this stage.
+                                dataset = filtered_datasets[tag]
+                                model.update(dataset)
+                                optimize_model_and_save_result(model, dataset)
 
                 if summary_writer:
                     with summary_writer.as_default(step=step):
@@ -750,12 +884,57 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
 
         tf.print("Optimization completed without errors", output_stream=absl.logging.INFO)
 
-        record = Record(datasets, models, acquisition_state)
+        record = Record(datasets, models, history_state)
         result = OptimizationResult(Ok(record), history)
         if track_state and track_path is not None:
             result.save_result(Path(track_path) / OptimizationResult.RESULTS_FILENAME)
         return result
 
+    def continue_optimization(
+        self,
+        num_steps: int,
+        optimization_result: OptimizationResult[StateType, TrainableProbabilisticModelType],
+        *args: Any,
+        **kwargs: Any,
+    ) -> OptimizationResult[StateType, TrainableProbabilisticModelType]:
+        """
+        Continue a previous optimization that either failed, was terminated early, or which
+        you simply wish to run for more steps.
+
+        :param num_steps: The total number of optimization steps, including any that have already
+            been run.
+        :param optimization_result: The optimization result from which to extract the datasets,
+            models and acquisition state. If the result was successful then the final result is
+            used; otherwise the last record in the history is used. The size of the history
+            is used to determine how many more steps are required.
+        :param args: Any more positional arguments to pass on to optimize.
+        :param kwargs: Any more keyword arguments to pass on to optimize.
+        :return: An :class:`OptimizationResult`. The history will contain both the history from
+            `optimization_result` (including the `final_result` if that was successful) and
+            any new records.
+        """
+        history: list[
+            Record[StateType, TrainableProbabilisticModelType]
+            | FrozenRecord[StateType, TrainableProbabilisticModelType]
+        ] = []
+        history.extend(optimization_result.history)
+        if optimization_result.final_result.is_ok:
+            history.append(optimization_result.final_result.unwrap())
+        if not history:
+            raise ValueError("Cannot continue from empty optimization result")
+
+        result = self.optimize(  # type: ignore[call-overload]
+            num_steps,
+            history[-1].datasets,
+            history[-1].models,
+            *args,
+            acquisition_state=history[-1].acquisition_state,
+            **kwargs,
+            start_step=len(history) - 1,
+        )
+        result.history[:1] = history
+        return result
+
 
 def write_summary_init(
     observer: Observer,
@@ -786,13 +965,17 @@ def write_summary_init(
 
 def write_summary_initial_model_fit(
     datasets: Mapping[Tag, Dataset],
-    models: Mapping[Tag, TrainableProbabilisticModel],
+    models: Mapping[Tag, ProbabilisticModel],
     model_fitting_timer: Timer,
 ) -> None:
     """Write TensorBoard summary for the model fitting to the initial data."""
     for tag, model in models.items():
         with tf.name_scope(f"{tag}.model"):
-            model.log(datasets[tag])
+            # Prefer local dataset if available.
+            tags = [tag, LocalizedTag.from_tag(tag).global_tag]
+            _, dataset = get_value_for_tag(datasets, *tags)
+            assert dataset is not None
+            model.log(dataset)
     logging.scalar(
         "wallclock/model_fitting",
         model_fitting_timer.time,
@@ -833,13 +1016,16 @@ def observation_plot_init(
 
 def write_summary_observations(
     datasets: Mapping[Tag, Dataset],
-    models: Mapping[Tag, TrainableProbabilisticModel],
+    models: Mapping[Tag, ProbabilisticModel],
     tagged_output: Mapping[Tag, TensorType],
     model_fitting_timer: Timer,
     observation_plot_dfs: MutableMapping[Tag, pd.DataFrame],
 ) -> None:
     """Write TensorBoard summary for the current step observations."""
-    for tag in datasets:
+    for tag in models:
+        if tag not in tagged_output:
+            continue
+
         with tf.name_scope(f"{tag}.model"):
             models[tag].log(datasets[tag])
 
@@ -930,7 +1116,7 @@ def write_summary_observations(
 
 def write_summary_query_points(
     datasets: Mapping[Tag, Dataset],
-    models: Mapping[Tag, TrainableProbabilisticModel],
+    models: Mapping[Tag, ProbabilisticModel],
     search_space: SearchSpace,
     query_points: TensorType,
     query_point_generation_timer: Timer,
@@ -941,7 +1127,7 @@ def write_summary_query_points(
     if tf.rank(query_points) == 2:
         for i in tf.range(tf.shape(query_points)[1]):
             if len(query_points) == 1:
-                logging.scalar(f"query_points/[{i}]", float(query_points[0, i]))
+                logging.scalar(f"query_point/[{i}]", float(query_points[0, i]))
             else:
                 logging.histogram(f"query_points/[{i}]", query_points[:, i])
         logging.histogram("query_points/euclidean_distances", lambda: pdist(query_points))
@@ -987,6 +1173,7 @@ def stop_at_minimum(
     minimizers_atol: float = 0,
     minimizers_rtol: float = 0.05,
     objective_tag: Tag = OBJECTIVE,
+    minimum_step_number: Optional[int] = None,
 ) -> EarlyStopCallback[TrainableProbabilisticModel, object]:
     """
     Generate an early stop function that terminates a BO loop when it gets close enough to the
@@ -999,6 +1186,7 @@ def stop_at_minimum(
     :param minimizers_atol: Absolute tolerance for minimizer point.
     :param minimizers_rtol: Relative tolerance for minimizer point.
     :param objective_tag: The tag for the objective data.
+    :param minimum_step_number: Minimum step number to stop at.
     :return: An early stop function that terminates if we get close enough to both the minimum
         and any of the minimizer points.
     """
@@ -1008,6 +1196,8 @@ def stop_at_minimum(
         _models: Mapping[Tag, TrainableProbabilisticModel],
         _acquisition_state: object,
     ) -> bool:
+        if minimum_step_number is not None and logging.get_step_number() < minimum_step_number:
+            return False
         dataset = datasets[objective_tag]
         arg_min_idx = tf.squeeze(tf.argmin(dataset.observations, axis=0))
         if minimum is not None:
