@@ -14,8 +14,10 @@
 
 from __future__ import annotations
 
-from typing import Callable, Optional, Sequence
+import io
+from typing import Callable, List, Optional, Sequence, Union
 
+import imageio
 import matplotlib.pyplot as plt
 import numpy as np
 import tensorflow as tf
@@ -23,17 +25,24 @@ from gpflow.models import GPModel
 from matplotlib import cm
 from matplotlib.axes import Axes
 from matplotlib.collections import Collection
+from matplotlib.colors import rgb2hex
 from matplotlib.contour import ContourSet
 from matplotlib.figure import Figure
+from matplotlib.patches import Rectangle
 
 from trieste.acquisition import AcquisitionFunction
 from trieste.acquisition.multi_objective.dominance import non_dominated
+from trieste.bayesian_optimizer import FrozenRecord, Record, StateType
+from trieste.models import ProbabilisticModel
+from trieste.observer import OBJECTIVE
+from trieste.space import TaggedMultiSearchSpace
 from trieste.types import TensorType
 from trieste.utils import to_numpy
+from trieste.utils.misc import LocalizedTag
 
 
 def create_grid(
-    mins: TensorType, maxs: TensorType, grid_density: int = 20
+    mins: TensorType, maxs: TensorType, grid_density: int = 30
 ) -> tuple[TensorType, TensorType, TensorType]:
     """
     Creates a regular 2D grid of size `grid_density^2` between mins and maxs.
@@ -97,13 +106,13 @@ def plot_function_2d(
     obj_func: Callable[[TensorType], TensorType],
     mins: TensorType,
     maxs: TensorType,
-    grid_density: int = 20,
+    grid_density: int = 100,
     contour: bool = False,
     log: bool = False,
     title: Optional[Sequence[str]] = None,
     xlabel: Optional[str] = None,
     ylabel: Optional[str] = None,
-    figsize: Optional[tuple[float, float]] = None,
+    figsize: Optional[tuple[float, float]] = (8, 6),
     colorbar: bool = False,
     alpha: float = 1.0,
     fill: bool = False,
@@ -175,13 +184,13 @@ def plot_acq_function_2d(
     acq_func: AcquisitionFunction,
     mins: TensorType,
     maxs: TensorType,
-    grid_density: int = 20,
+    grid_density: int = 100,
     contour: bool = False,
     log: bool = False,
     title: Optional[Sequence[str]] = None,
     xlabel: Optional[str] = None,
     ylabel: Optional[str] = None,
-    figsize: Optional[tuple[float, float]] = None,
+    figsize: Optional[tuple[float, float]] = (8, 6),
     colorbar: bool = False,
     alpha: float = 1.0,
     fill: bool = False,
@@ -228,21 +237,21 @@ def plot_acq_function_2d(
 
 def format_point_markers(
     num_pts: int,
-    num_init: Optional[int] = None,
-    idx_best: Optional[int] = None,
+    num_init: Optional[Union[int, TensorType]] = None,
+    idx_best: Optional[TensorType] = None,
     mask_fail: Optional[TensorType] = None,
     m_init: str = "x",
     m_add: str = "o",
     c_pass: str = "tab:green",
-    c_fail: str = "tab:red",
+    c_fail: Union[str, List[str]] = "tab:red",
     c_best: str = "tab:purple",
 ) -> tuple[TensorType, TensorType]:
     """
     Prepares point marker styles according to some BO factors.
 
     :param num_pts: total number of BO points
-    :param num_init: initial number of BO points
-    :param idx_best: index of the best BO point
+    :param num_init: initial number of BO points; can also be a mask
+    :param idx_best: index of the best BO point(s)
     :param mask_fail: Bool vector, True if the corresponding observation violates the constraint(s)
     :param m_init: marker for the initial BO points
     :param m_add: marker for the other BO points
@@ -256,7 +265,10 @@ def format_point_markers(
     col_pts = np.repeat(c_pass, num_pts)
     col_pts = col_pts.astype("<U15")
     mark_pts = np.repeat(m_init, num_pts)
-    mark_pts[num_init:] = m_add
+    if isinstance(num_init, int):
+        mark_pts[num_init:] = m_add
+    else:
+        mark_pts[np.where(~num_init)] = m_add
     if mask_fail is not None:
         col_pts[np.where(mask_fail)] = c_fail
     if idx_best is not None:
@@ -268,14 +280,14 @@ def format_point_markers(
 def plot_bo_points(
     pts: TensorType,
     ax: Axes,
-    num_init: Optional[int] = None,
+    num_init: Optional[Union[int, TensorType]] = None,
     idx_best: Optional[int] = None,
     mask_fail: Optional[TensorType] = None,
     obs_values: Optional[TensorType] = None,
     m_init: str = "x",
     m_add: str = "o",
     c_pass: str = "tab:green",
-    c_fail: str = "tab:red",
+    c_fail: Union[str, List[str]] = "tab:red",
     c_best: str = "tab:purple",
 ) -> None:
     """
@@ -284,7 +296,7 @@ def plot_bo_points(
 
     :param pts: [N, 2] x inputs
     :param ax: a plt axes object
-    :param num_init: initial number of BO points
+    :param num_init: initial number of BO points; can also be a mask
     :param idx_best: index of the best BO point
     :param mask_fail: Bool vector, True if the corresponding observation violates the constraint(s)
     :param obs_values: optional [N] outputs (for 3d plots)
@@ -313,7 +325,7 @@ def plot_mobo_points_in_obj_space(
     obs_values: TensorType,
     num_init: Optional[int] = None,
     mask_fail: Optional[TensorType] = None,
-    figsize: Optional[tuple[float, float]] = None,
+    figsize: Optional[tuple[float, float]] = (8, 6),
     xlabel: str = "Obj 1",
     ylabel: str = "Obj 2",
     zlabel: str = "Obj 3",
@@ -347,7 +359,7 @@ def plot_mobo_points_in_obj_space(
     """
     obj_num = obs_values.shape[-1]
     tf.debugging.assert_shapes([])
-    assert obj_num == 2 or obj_num == 3, NotImplementedError(
+    assert obj_num in (2, 3), NotImplementedError(
         f"Only support 2/3-objective functions but found: {obj_num}"
     )
 
@@ -386,7 +398,7 @@ def plot_mobo_history(
     metric_func: Callable[[TensorType], TensorType],
     num_init: int,
     mask_fail: Optional[TensorType] = None,
-    figsize: Optional[tuple[float, float]] = None,
+    figsize: Optional[tuple[float, float]] = (8, 6),
 ) -> tuple[Figure, Axes]:
     """
     Draw the performance measure for multi-objective optimization.
@@ -461,11 +473,11 @@ def plot_gp_2d(
     model: GPModel,
     mins: TensorType,
     maxs: TensorType,
-    grid_density: int = 20,
+    grid_density: int = 100,
     contour: bool = False,
     xlabel: Optional[str] = None,
     ylabel: Optional[str] = None,
-    figsize: Optional[tuple[float, float]] = None,
+    figsize: Optional[tuple[float, float]] = (8, 6),
     predict_y: bool = False,
 ) -> tuple[Figure, Axes]:
     """
@@ -534,3 +546,152 @@ def plot_gp_2d(
             axx.set_ylim(mins[1], maxs[1])
 
     return fig, ax
+
+
+def plot_trust_region_history_2d(
+    obj_func: Callable[[TensorType], TensorType],
+    mins: TensorType,
+    maxs: TensorType,
+    history: Record[StateType, ProbabilisticModel] | FrozenRecord[StateType, ProbabilisticModel],
+    num_query_points: Optional[int] = None,
+    num_init: Optional[int] = None,
+    alpha: float = 0.3,
+) -> tuple[Optional[Figure], Optional[Axes]]:
+    """
+    Plot the contour of the objective function, query points and the trust regions for a particular
+    step of the optimization process.
+
+    :param obj_func: the objective function that returns a n-array given a [n, d] array
+    :param mins: search space 2D lower bounds
+    :param maxs: search space 2D upper bounds
+    :param history: the optimization history for a particular step of the optimization process
+    :param num_query_points: total number of query points in this step
+    :param num_init: initial number of BO points
+    :param alpha: transparency for the trust regions
+    :return: figure and axes
+    """
+
+    state = history.acquisition_state
+    if state is None:
+        # If state is None, then there is no trust region state to plot.
+        return None, None
+
+    # Plot objective contour.
+    fig, ax = plot_function_2d(
+        obj_func,
+        mins,
+        maxs,
+        grid_density=40,
+        contour=True,
+    )
+
+    assert hasattr(state, "acquisition_space")
+    acquisition_space = state.acquisition_space
+
+    if isinstance(acquisition_space, TaggedMultiSearchSpace):
+        spaces = [acquisition_space.get_subspace(tag) for tag in acquisition_space.subspace_tags]
+    else:
+        spaces = [acquisition_space]
+
+    num_spaces = len(spaces)
+    if num_query_points is None:
+        num_query_points = num_spaces  # Assume one query point per subspace.
+
+    assert num_query_points % num_spaces == 0, (
+        f"Number of query points ({num_query_points}) must be a multiple of the number of "
+        f"spaces ({num_spaces})."
+    )
+    num_query_points_per_space = num_query_points // num_spaces
+
+    # If there are local datasets, use them to generate the colors for the query points.
+    # Otherwise, use the global dataset and assume the last `num_query_points` points are new.
+    if len(history.datasets) > 1:
+        # Expect there to be an objective dataset for each subspace.
+        datasets = [history.datasets[LocalizedTag(OBJECTIVE, i)] for i in range(num_spaces)]
+
+        _new_points_mask = [
+            np.zeros(dataset.query_points.shape[0], dtype=bool) for dataset in datasets
+        ]
+        # New points in each dataset are at the end.
+        for mask in _new_points_mask:
+            mask[-num_query_points_per_space:] = True
+        # Concatenate the masks.
+        new_points_mask = np.concatenate(_new_points_mask)
+
+        if num_init is not None:
+            _num_init_mask = [
+                np.zeros(dataset.query_points.shape[0], dtype=bool) for dataset in datasets
+            ]
+            # First num_init points in each dataset are the init points.
+            for mask in _num_init_mask:
+                mask[:num_init] = True
+            # Concatenate the masks.
+            num_init = np.concatenate(_num_init_mask)
+
+        # Get the overall query points.
+        query_points = np.concatenate([dataset.query_points for dataset in datasets])
+    else:
+        query_points = history.dataset.query_points  # All query points.
+        new_points_mask = np.zeros(query_points.shape[0], dtype=bool)
+        new_points_mask[-num_query_points:] = True
+
+    # Plot trust regions.
+    space_colors = [rgb2hex(color) for color in cm.rainbow(np.linspace(0, 1, num_spaces))]
+    for i, space in enumerate(spaces):
+        lb = space.lower
+        ub = space.upper
+        ax[0, 0].add_patch(
+            Rectangle(
+                (lb[0], lb[1]),
+                ub[0] - lb[0],
+                ub[1] - lb[1],
+                facecolor=space_colors[i],
+                edgecolor=space_colors[i],
+                alpha=alpha,
+            )
+        )
+
+    # Plot new query points, using failure mask to color them.
+    point_colors = (
+        np.reshape(
+            space_colors * num_query_points_per_space, [num_query_points_per_space, num_spaces]
+        )
+        .flatten("F")
+        .tolist()
+    )
+    plot_bo_points(
+        query_points,
+        ax[0, 0],
+        num_init,
+        mask_fail=new_points_mask,
+        c_pass="black",
+        c_fail=point_colors,
+    )
+
+    return fig, ax
+
+
+def convert_figure_to_frame(fig: plt.Figure) -> TensorType:
+    """
+    Converts a matplotlib figure to an array of pixels.
+
+    :param fig: a matplotlib figure
+    :return: an array of pixels - a frame
+    """
+    fig.canvas.draw()
+    size_pix = fig.get_size_inches() * fig.dpi
+    image = np.frombuffer(fig.canvas.tostring_rgb(), dtype="uint8")
+    return image.reshape(list(size_pix[::-1].astype(int)) + [3])
+
+
+def convert_frames_to_gif(frames: Sequence[TensorType], duration: int = 5000) -> io.BytesIO:
+    """
+    Converts a sequence of frames (arrays of pixels) to a gif.
+
+    :param frames: sequence of frames
+    :param duration: duration of each frame in milliseconds
+    :return: gif file
+    """
+    gif_file = io.BytesIO()
+    imageio.mimsave(gif_file, frames, format="gif", loop=0, duration=duration)
+    return gif_file
