@@ -34,43 +34,46 @@ from ...space import Box, SearchSpace
 from ...types import TensorType
 from ..gpflow.models import GaussianProcessRegression
 
+# NOTE: As a static non-Tensor, this should really be a tf.constant (like the other constants).
+# However, changing it breaks serialisation during the expected_improvement.pct.py notebook.
+# See https://github.com/secondmind-labs/trieste/issues/833 for details.
 KERNEL_LENGTHSCALE = tf.cast(0.2, dtype=gpflow.default_float())
 """
 Default value of the kernel lengthscale parameter.
 """
 
 
-KERNEL_PRIOR_SCALE = tf.cast(1.0, dtype=gpflow.default_float())
+KERNEL_PRIOR_SCALE = tf.constant(1.0, dtype=gpflow.default_float())
 """
 Default value of the scaling factor for the kernel lengthscale and variance parameters.
 """
 
 
-CLASSIFICATION_KERNEL_VARIANCE_NOISE_FREE = tf.cast(100.0, dtype=gpflow.default_float())
+CLASSIFICATION_KERNEL_VARIANCE_NOISE_FREE = tf.constant(100.0, dtype=gpflow.default_float())
 """
 Default value of the kernel variance parameter for classification models in the noise free case.
 """
 
 
-CLASSIFICATION_KERNEL_VARIANCE = tf.cast(1.0, dtype=gpflow.default_float())
+CLASSIFICATION_KERNEL_VARIANCE = tf.constant(1.0, dtype=gpflow.default_float())
 """
 Default value of the kernel variance parameter for classification models.
 """
 
 
-MAX_NUM_INDUCING_POINTS = tf.cast(500, dtype=tf.int32)
+MAX_NUM_INDUCING_POINTS = tf.constant(500, dtype=tf.int32)
 """
 Default maximum number of inducing points.
 """
 
 
-NUM_INDUCING_POINTS_PER_DIM = tf.cast(25, dtype=tf.int32)
+NUM_INDUCING_POINTS_PER_DIM = tf.constant(25, dtype=tf.int32)
 """
 Default number of inducing points per dimension of the search space.
 """
 
 
-SIGNAL_NOISE_RATIO_LIKELIHOOD = tf.cast(10, dtype=gpflow.default_float())
+SIGNAL_NOISE_RATIO_LIKELIHOOD = tf.constant(10, dtype=gpflow.default_float())
 """
 Default value used for initializing (noise) variance parameter of the likelihood function.
 If user does not specify it, the noise variance is set to maintain the signal to noise ratio
@@ -119,12 +122,12 @@ def build_gpr(
     """
     empirical_mean, empirical_variance, _ = _get_data_stats(data)
 
-    if kernel is None and search_space is None:
-        raise ValueError(
-            "'build_gpr' function requires one of 'search_space' or 'kernel' arguments,"
-            " but got neither"
-        )
-    elif kernel is None and search_space is not None:
+    if kernel is None:
+        if search_space is None:
+            raise ValueError(
+                "'build_gpr' function requires one of 'search_space' or 'kernel' arguments,"
+                " but got neither"
+            )
         kernel = _get_kernel(empirical_variance, search_space, kernel_priors, kernel_priors)
     mean = _get_mean_function(empirical_mean)
 
@@ -264,7 +267,7 @@ def build_vgp_classifier(
 
     model_likelihood = gpflow.likelihoods.Bernoulli()
     kernel = _get_kernel(variance, search_space, kernel_priors, add_prior_to_variance)
-    mean = _get_mean_function(tf.cast(0.0, dtype=gpflow.default_float()))
+    mean = _get_mean_function(tf.constant(0.0, dtype=gpflow.default_float()))
 
     model = VGP(data.astuple(), kernel, model_likelihood, mean_function=mean)
 
@@ -336,7 +339,7 @@ def build_svgp(
 
     if classification:
         empirical_variance = CLASSIFICATION_KERNEL_VARIANCE
-        empirical_mean = tf.cast(0.0, dtype=gpflow.default_float())
+        empirical_mean = tf.constant(0.0, dtype=gpflow.default_float())
         model_likelihood = gpflow.likelihoods.Bernoulli()
     else:
         model_likelihood = gpflow.likelihoods.Gaussian()
@@ -400,7 +403,7 @@ def _get_lengthscales(search_space: SearchSpace) -> TensorType:
     )
     search_space_collapsed = tf.equal(search_space.upper, search_space.lower)
     lengthscales = tf.where(
-        search_space_collapsed, tf.cast(1.0, dtype=gpflow.default_float()), lengthscales
+        search_space_collapsed, tf.constant(1.0, dtype=gpflow.default_float()), lengthscales
     )
     return lengthscales
 
@@ -589,7 +592,7 @@ def _create_multifidelity_nonlinear_autoregressive_kernels(
     scale_lengthscale = 1.0
     kernels = [kernel_base_class(lengthscales=lengthscales)]
 
-    for i in range(1, n_fidelities):
+    for _ in range(1, n_fidelities):
         interaction_kernel = kernel_base_class(lengthscales=lengthscales, active_dims=dims[:-1])
         scale_kernel = kernel_base_class(lengthscales=scale_lengthscale, active_dims=[dims[-1]])
         bias_kernel = kernel_base_class(lengthscales=lengthscales, active_dims=dims[:-1])
@@ -609,10 +612,10 @@ def _create_multifidelity_nonlinear_autoregressive_kernels(
 
         if add_prior_to_variance:
             interaction_kernel.variance.prior = tfp.distributions.LogNormal(
-                tf.cast(0.0, dtype=gpflow.default_float()), KERNEL_PRIOR_SCALE
+                tf.constant(0.0, dtype=gpflow.default_float()), KERNEL_PRIOR_SCALE
             )
             bias_kernel.variance.prior = tfp.distributions.LogNormal(
-                tf.cast(0.0, dtype=gpflow.default_float()), KERNEL_PRIOR_SCALE
+                tf.constant(0.0, dtype=gpflow.default_float()), KERNEL_PRIOR_SCALE
             )
 
         kernels.append(interaction_kernel * scale_kernel + bias_kernel)
