@@ -19,6 +19,7 @@ from typing import Optional, Sequence, Tuple, Union, cast
 import gpflow
 import tensorflow as tf
 import tensorflow_probability as tfp
+from check_shapes import inherit_check_shapes
 from gpflow.conditionals.util import sample_mvn
 from gpflow.inducing_variables import (
     SeparateIndependentInducingVariables,
@@ -45,10 +46,11 @@ from ..interfaces import (
     SupportsCovarianceWithTopFidelity,
     SupportsGetInducingVariables,
     SupportsGetInternalData,
+    SupportsPredictY,
     TrainableProbabilisticModel,
     TrajectorySampler,
 )
-from ..optimizer import BatchOptimizer, Optimizer
+from ..optimizer import BatchOptimizer, Optimizer, OptimizeResult
 from .inducing_point_selectors import InducingPointSelector
 from .interface import GPflowPredictor, SupportsCovarianceBetweenPoints
 from .sampler import DecoupledTrajectorySampler, RandomFourierFeatureTrajectorySampler
@@ -64,7 +66,6 @@ from .utils import (
 
 class GaussianProcessRegression(
     GPflowPredictor,
-    TrainableProbabilisticModel,
     FastUpdateModel,
     SupportsCovarianceBetweenPoints,
     SupportsGetInternalData,
@@ -158,6 +159,7 @@ class GaussianProcessRegression(
             ),
         )
 
+    @inherit_check_shapes
     def predict_y(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
         f_mean, f_var = self.predict(query_points)
         return self.model.likelihood.predict_mean_and_var(query_points, f_mean, f_var)
@@ -247,7 +249,7 @@ class GaussianProcessRegression(
 
         return cov
 
-    def optimize(self, dataset: Dataset) -> None:
+    def optimize(self, dataset: Dataset) -> OptimizeResult:
         """
         Optimize the model with the specified `dataset`.
 
@@ -282,8 +284,9 @@ class GaussianProcessRegression(
             self.find_best_model_initialization(
                 self._num_kernel_samples * num_trainable_params_with_priors_or_constraints
             )
-        self.optimizer.optimize(self.model, dataset)
+        result = self.optimizer.optimize(self.model, dataset)
         self.update_posterior_cache()
+        return result
 
     def find_best_model_initialization(self, num_kernel_samples: int) -> None:
         """
@@ -521,7 +524,6 @@ class GaussianProcessRegression(
 
 class SparseGaussianProcessRegression(
     GPflowPredictor,
-    TrainableProbabilisticModel,
     SupportsCovarianceBetweenPoints,
     SupportsGetInducingVariables,
     SupportsGetInternalData,
@@ -607,6 +609,7 @@ class SparseGaussianProcessRegression(
     ) -> Optional[InducingPointSelector[SparseGaussianProcessRegression]]:
         return self._inducing_point_selector
 
+    @inherit_check_shapes
     def predict_y(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
         f_mean, f_var = self.predict(query_points)
         return self.model.likelihood.predict_mean_and_var(query_points, f_mean, f_var)
@@ -634,14 +637,15 @@ class SparseGaussianProcessRegression(
         if not is_variable(self._model.num_data):
             self._model.num_data = tf.Variable(self._model.num_data, trainable=False)
 
-    def optimize(self, dataset: Dataset) -> None:
+    def optimize(self, dataset: Dataset) -> OptimizeResult:
         """
         Optimize the model with the specified `dataset`.
 
         :param dataset: The data with which to optimize the `model`.
         """
-        self.optimizer.optimize(self.model, dataset)
+        result = self.optimizer.optimize(self.model, dataset)
         self.update_posterior_cache()
+        return result
 
     def update(self, dataset: Dataset) -> None:
         self._ensure_variable_model_data()
@@ -833,7 +837,6 @@ class SparseGaussianProcessRegression(
 
 class SparseVariational(
     GPflowPredictor,
-    TrainableProbabilisticModel,
     SupportsCovarianceBetweenPoints,
     SupportsGetInducingVariables,
     HasTrajectorySampler,
@@ -929,6 +932,7 @@ class SparseVariational(
     def inducing_point_selector(self) -> Optional[InducingPointSelector[SparseVariational]]:
         return self._inducing_point_selector
 
+    @inherit_check_shapes
     def predict_y(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
         f_mean, f_var = self.predict(query_points)
         return self.model.likelihood.predict_mean_and_var(query_points, f_mean, f_var)
@@ -975,14 +979,15 @@ class SparseVariational(
                 self._update_inducing_variables(new_inducing_points)
                 self.update_posterior_cache()
 
-    def optimize(self, dataset: Dataset) -> None:
+    def optimize(self, dataset: Dataset) -> OptimizeResult:
         """
         Optimize the model with the specified `dataset`.
 
         :param dataset: The data with which to optimize the `model`.
         """
-        self.optimizer.optimize(self.model, dataset)
+        result = self.optimizer.optimize(self.model, dataset)
         self.update_posterior_cache()
+        return result
 
     def _update_inducing_variables(self, new_inducing_points: TensorType) -> None:
         """
@@ -1087,26 +1092,13 @@ class SparseVariational(
         Return a trajectory sampler. For :class:`SparseVariational`, we build
         trajectories using a decoupled random Fourier feature approximation.
 
-        At the moment only models with single latent GP are supported.
-
         :return: The trajectory sampler.
-        :raise NotImplementedError: If we try to use the
-            sampler with a model that has more than one latent GP.
         """
-        if self.model.num_latent_gps > 1:
-            raise NotImplementedError(
-                f"""
-                Trajectory sampler does not currently support models with multiple latent
-                GPs, however received a model with {self.model.num_latent_gps} latent GPs.
-                """
-            )
-
         return DecoupledTrajectorySampler(self, self._num_rff_features)
 
 
 class VariationalGaussianProcess(
     GPflowPredictor,
-    TrainableProbabilisticModel,
     SupportsCovarianceBetweenPoints,
     SupportsGetInducingVariables,
     HasTrajectorySampler,
@@ -1228,6 +1220,10 @@ class VariationalGaussianProcess(
                 ),
             )
 
+            # reinitialise the model so that the underlying Parameters have the right shape
+            # and then reassign the original values
+            old_q_mu = model.q_mu
+            old_q_sqrt = model.q_sqrt
             model.__init__(  # type: ignore[misc]
                 variable_data,
                 model.kernel,
@@ -1235,6 +1231,8 @@ class VariationalGaussianProcess(
                 model.mean_function,
                 model.num_latent_gps,
             )
+            model.q_mu.assign(old_q_mu)
+            model.q_sqrt.assign(old_q_sqrt)
 
     def __repr__(self) -> str:
         """"""
@@ -1247,6 +1245,7 @@ class VariationalGaussianProcess(
     def model(self) -> VGP:
         return self._model
 
+    @inherit_check_shapes
     def predict_y(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
         f_mean, f_var = self.predict(query_points)
         return self.model.likelihood.predict_mean_and_var(query_points, f_mean, f_var)
@@ -1263,7 +1262,7 @@ class VariationalGaussianProcess(
         update_vgp_data(self.model, (dataset.query_points, dataset.observations))
         self.update_posterior_cache()
 
-    def optimize(self, dataset: Dataset) -> None:
+    def optimize(self, dataset: Dataset) -> Optional[OptimizeResult]:
         """
         :class:`VariationalGaussianProcess` has a custom `optimize` method that (optionally) permits
         alternating between standard optimization steps (for kernel parameters) and natural gradient
@@ -1298,13 +1297,14 @@ class VariationalGaussianProcess(
             for _ in range(base_optimizer.max_iter):  # type: ignore
                 perform_optimization_step()
 
-            gpflow.set_trainable(model.q_mu, True)  # revert varitional params to trainable
+            gpflow.set_trainable(model.q_mu, True)  # revert variational params to trainable
             gpflow.set_trainable(model.q_sqrt, True)
-
+            result = None  # TODO: find something useful to return
         else:
-            self.optimizer.optimize(model, dataset)
+            result = self.optimizer.optimize(model, dataset)
 
         self.update_posterior_cache()
+        return result
 
     def get_inducing_variables(self) -> Tuple[TensorType, TensorType, TensorType, bool]:
         """
@@ -1358,7 +1358,7 @@ class VariationalGaussianProcess(
             (L being the number of latent GPs = number of output dimensions)
         """
 
-        inducing_points, _, q_sqrt, whiten = self.get_inducing_variables()
+        _, _, q_sqrt, whiten = self.get_inducing_variables()
 
         return _covariance_between_points_for_variational_models(
             kernel=self.get_kernel(),
@@ -1370,7 +1370,9 @@ class VariationalGaussianProcess(
         )
 
 
-class MultifidelityAutoregressive(TrainableProbabilisticModel, SupportsCovarianceWithTopFidelity):
+class MultifidelityAutoregressive(
+    TrainableProbabilisticModel, SupportsPredictY, SupportsCovarianceWithTopFidelity
+):
     r"""
     A :class:`TrainableProbabilisticModel` implementation of the model
     from :cite:`Kennedy2000`. This is a multi-fidelity model that works with an
@@ -1415,6 +1417,7 @@ class MultifidelityAutoregressive(TrainableProbabilisticModel, SupportsCovarianc
     def num_fidelities(self) -> int:
         return self._num_fidelities
 
+    @inherit_check_shapes
     def predict(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
         """
         Predict the marginal mean and variance at query_points.
@@ -1495,6 +1498,7 @@ class MultifidelityAutoregressive(TrainableProbabilisticModel, SupportsCovarianc
         )
         return residuals
 
+    @inherit_check_shapes
     def sample(self, query_points: TensorType, num_samples: int) -> TensorType:
         """
         Sample `num_samples` samples from the posterior distribution at `query_points`
@@ -1525,12 +1529,13 @@ class MultifidelityAutoregressive(TrainableProbabilisticModel, SupportsCovarianc
             )  # [S, N, P]
 
             mask = query_points_fidelity_col >= fidelity  # [N, P]
-            mask = tf.repeat(mask[..., None, :, :], num_samples, axis=-3)
+            mask = tf.broadcast_to(mask[..., None, :, :], new_fidelity_signal_sample.shape)
 
             signal_sample = tf.where(mask, new_fidelity_signal_sample, signal_sample)
 
         return signal_sample
 
+    @inherit_check_shapes
     def predict_y(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
         """
         Predict the marginal mean and variance at `query_points` including observation noise
@@ -1654,9 +1659,12 @@ class MultifidelityAutoregressive(TrainableProbabilisticModel, SupportsCovarianc
 
         return f_var
 
+    def log(self, dataset: Optional[Dataset] = None) -> None:
+        return
+
 
 class MultifidelityNonlinearAutoregressive(
-    TrainableProbabilisticModel, SupportsCovarianceWithTopFidelity
+    TrainableProbabilisticModel, SupportsPredictY, SupportsCovarianceWithTopFidelity
 ):
     r"""
     A :class:`TrainableProbabilisticModel` implementation of the model from
@@ -1697,6 +1705,7 @@ class MultifidelityNonlinearAutoregressive(
     def num_fidelities(self) -> int:
         return self._num_fidelities
 
+    @inherit_check_shapes
     def sample(self, query_points: TensorType, num_samples: int) -> TensorType:
         """
         Return ``num_samples`` samples from the independent marginal distributions at
@@ -1712,18 +1721,20 @@ class MultifidelityNonlinearAutoregressive(
         ) = check_and_extract_fidelity_query_points(
             query_points, max_fidelity=self.num_fidelities - 1
         )  # [..., N, D], [..., N, 1]
-        # Repeat query_points to get same shape as signal sample
-        query_points_fidelity_col = tf.repeat(
-            query_points_fidelity_col[..., None, :, :], num_samples, axis=-3
-        )  # [..., S, N, 1]
 
         signal_sample = self.fidelity_models[0].sample(
             query_points_wo_fidelity, num_samples
         )  # [..., S, N, 1]
 
+        # Repeat query_points to get same shape as signal sample
+        query_points_fidelity_col = tf.broadcast_to(
+            query_points_fidelity_col[..., None, :, :], signal_sample.shape
+        )  # [..., S, N, 1]
+
         for fidelity in range(1, self.num_fidelities):
-            qp_repeated = tf.repeat(
-                query_points_wo_fidelity[..., None, :, :], num_samples, axis=-3
+            qp_repeated = tf.broadcast_to(
+                query_points_wo_fidelity[..., None, :, :],
+                signal_sample.shape[:-1] + query_points_wo_fidelity.shape[-1],
             )  # [..., S, N, D]
             qp_augmented = tf.concat([qp_repeated, signal_sample], axis=-1)  # [..., S, N, D + 1]
             new_signal_sample = self.fidelity_models[fidelity].sample(
@@ -1738,6 +1749,7 @@ class MultifidelityNonlinearAutoregressive(
 
         return signal_sample
 
+    @inherit_check_shapes
     def predict(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
         """
         Predict the marginal mean and variance at query_points.
@@ -1758,6 +1770,7 @@ class MultifidelityNonlinearAutoregressive(
         mean = tf.reduce_mean(sample_mean, axis=-1)
         return mean, variance
 
+    @inherit_check_shapes
     def predict_y(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
         """
         Predict the marginal mean and variance at `query_points` including observation noise
@@ -1814,16 +1827,21 @@ class MultifidelityNonlinearAutoregressive(
         # Create new dimension to store samples for each query point
         # Repeat the inital sample mean and variance S times and add a dimension in the
         # middle (so that sample mean and query points can be concatenated sensibly)
-        sample_mean = tf.repeat(sample_mean, self.monte_carlo_random_numbers.shape[0], axis=-1)[
+        sample_mean = tf.broadcast_to(
+            sample_mean, sample_mean.shape[:-1] + self.monte_carlo_random_numbers.shape[0]
+        )[
             ..., :, None, :
         ]  # [..., N, 1, S]
-        sample_var = tf.repeat(sample_var, self.monte_carlo_random_numbers.shape[0], axis=-1)[
+        sample_var = tf.broadcast_to(
+            sample_var, sample_var.shape[:-1] + self.monte_carlo_random_numbers.shape[0]
+        )[
             ..., :, None, :
         ]  # [..., N, 1, S]
 
         # Repeat fidelity points for each sample to match shapes for masking
-        query_points_fidelity_col = tf.repeat(
-            query_points_fidelity_col, self.monte_carlo_random_numbers.shape[0], axis=-1
+        query_points_fidelity_col = tf.broadcast_to(
+            query_points_fidelity_col,
+            query_points_fidelity_col.shape[:-1] + self.monte_carlo_random_numbers.shape[0],
         )[
             ..., :, None, :
         ]  # [..., N, 1, S]
@@ -1876,17 +1894,16 @@ class MultifidelityNonlinearAutoregressive(
         # for concatentation with query points This also means that it has the same
         # shape as sample_var and sample_mean, so there's no broadcasting required
         # Note: at the moment we use the same monte carlo values for every value in the batch dim
-        reshaped_random_numbers = tf.repeat(
-            tf.transpose(self.monte_carlo_random_numbers), query_point.shape[-2], axis=-2
-        )[
-            ..., None, :
-        ]  # [..., N, 1, S]
+
+        reshaped_random_numbers = tf.broadcast_to(
+            tf.transpose(self.monte_carlo_random_numbers)[..., None, :],
+            sample_mean.shape,
+        )  # [..., N, 1, S]
         samples = reshaped_random_numbers * tf.sqrt(sample_var) + sample_mean  # [..., N, 1, S]
         # Add an extra unit dim to query_point and repeat for each of the samples
-        qp_repeated = tf.repeat(
+        qp_repeated = tf.broadcast_to(
             query_point[..., :, :, None],  # [..., N, D, 1]
-            self.monte_carlo_random_numbers.shape[0],
-            axis=-1,
+            query_point.shape + samples.shape[-1],
         )  # [..., N, D, S]
         qp_augmented = tf.concat([qp_repeated, samples], axis=-2)  # [..., N, D+1, S]
 
@@ -1984,21 +2001,23 @@ class MultifidelityNonlinearAutoregressive(
         ) = check_and_extract_fidelity_query_points(
             query_points, max_fidelity=self.num_fidelities - 1
         )  # [N, D], [N, 1]
-        # Repeat query_points to get same shape as signal sample
-        query_points_fidelity_col = tf.repeat(
-            query_points_fidelity_col[None, :, :], num_samples, axis=0
-        )
 
         # Signal sample stops updating once fidelity is reached for that query point
         signal_sample = self.fidelity_models[0].model.predict_f_samples(
             query_points_wo_fidelity, num_samples, full_cov=False
         )
+
+        # Repeat query_points to get same shape as signal sample
+        query_points_fidelity_col = tf.broadcast_to(
+            query_points_fidelity_col[None, :, :], signal_sample.shape
+        )
         # Max fidelity sample keeps updating to the max fidelity
         max_fidelity_sample = tf.identity(signal_sample)
 
         for fidelity in range(1, self.num_fidelities):
-            qp_repeated = tf.repeat(
-                query_points_wo_fidelity[None, :, :], num_samples, axis=0
+            qp_repeated = tf.broadcast_to(
+                query_points_wo_fidelity[None, :, :],
+                tf.TensorShape(num_samples) + query_points_wo_fidelity.shape,
             )  # [S, N, D]
             # We use max fidelity sample here, which is okay because anything
             # with a lower fidelity will not be updated
@@ -2017,3 +2036,6 @@ class MultifidelityNonlinearAutoregressive(
         cov = tfp.stats.covariance(signal_sample, max_fidelity_sample)[:, :, 0]
 
         return cov
+
+    def log(self, dataset: Optional[Dataset] = None) -> None:
+        return
