@@ -52,13 +52,29 @@ def assert_data_is_compatible(new_data: Dataset, existing_data: Dataset) -> None
 
 def randomize_hyperparameters(object: gpflow.Module) -> None:
     """
-    Sets hyperparameters to random samples from their constrained domains or (if not constraints
-    are available) their prior distributions.
+    Sets hyperparameters to random samples from their prior distributions or (for Sigmoid
+    constraints with no priors) their constrained domains. Note that it is up to the caller
+    to ensure that the prior, if defined, is compatible with the transform.
 
     :param object: Any gpflow Module.
     """
     for param in object.trainable_parameters:
-        if isinstance(param.bijector, tfp.bijectors.Sigmoid):
+        if param.prior is not None:
+            # handle constant priors for multi-dimensional parameters
+            # Use python conditionals here to avoid creating tensorflow `tf.cond` ops,
+            # i.e. using `len(param.shape)` instead of `tf.rank(param)`.
+            # Otherwise, tensorflow generates repeating random sequences for hyperparameters, see
+            # https://github.com/tensorflow/tensorflow/issues/61912.
+            if param.prior.batch_shape == param.prior.event_shape == [] and len(param.shape) == 1:
+                sample = param.prior.sample(tf.shape(param))
+            else:
+                sample = param.prior.sample()
+            if param.prior_on is gpflow.base.PriorOn.UNCONSTRAINED:
+                param.unconstrained_variable.assign(sample)
+            else:
+                param.assign(sample)
+
+        elif isinstance(param.bijector, tfp.bijectors.Sigmoid):
             sample = tf.random.uniform(
                 param.bijector.low.shape,
                 minval=param.bijector.low,
@@ -66,13 +82,6 @@ def randomize_hyperparameters(object: gpflow.Module) -> None:
                 dtype=param.bijector.low.dtype,
             )
             param.assign(sample)
-        elif param.prior is not None:
-            # handle constant priors for multi-dimensional parameters
-            if param.prior.batch_shape == param.prior.event_shape == [] and tf.rank(param) == 1:
-                sample = param.prior.sample(tf.shape(param))
-            else:
-                sample = param.prior.sample()
-            param.assign(sample)
 
 
 def squeeze_hyperparameters(
@@ -89,10 +98,10 @@ def squeeze_hyperparameters(
     :raise ValueError: If ``alpha`` is not in (0,1) or epsilon <= 0
     """
 
-    if not (0 < alpha < 1):
+    if not 0 < alpha < 1:
         raise ValueError(f"squeeze factor alpha must be in (0, 1), found {alpha}")
 
-    if not (0 < epsilon):
+    if epsilon <= 0:
         raise ValueError(f"offset factor epsilon must be > 0, found {epsilon}")
 
     for param in object.trainable_parameters:
@@ -299,7 +308,7 @@ def _compute_kernel_blocks(
         K12 = tf.repeat(tf.expand_dims(K12, -3), num_latent, axis=-3)
     elif len(tf.shape(K)) > 3:
         raise NotImplementedError(
-            "Covariance between points is not supported " "for kernels of type " f"{type(kernel)}."
+            "Covariance between points is not supported for kernels of type {type(kernel)}."
         )
 
     tf.debugging.assert_shapes(
