@@ -14,11 +14,13 @@
 
 from __future__ import annotations
 
-from typing import Any, Callable, Optional
+from typing import Any, Callable, Mapping, Optional
 
 import dill
 import gpflow
+import keras.callbacks
 import tensorflow as tf
+from check_shapes import inherit_check_shapes
 from gpflow.inducing_variables import InducingPoints
 from gpflux.layers import GPLayer, LatentVariableLayer
 from gpflux.models import DeepGP
@@ -62,6 +64,7 @@ class DeepGaussianProcess(
         optimizer: KerasOptimizer | None = None,
         num_rff_features: int = 1000,
         continuous_optimisation: bool = True,
+        compile_args: Optional[Mapping[str, Any]] = None,
     ):
         """
         :param model: The underlying GPflux deep Gaussian process model. Passing in a named closure
@@ -80,9 +83,23 @@ class DeepGaussianProcess(
         :param continuous_optimisation: if True (default), the optimizer will keep track of the
             number of epochs across BO iterations and use this number as initial_epoch. This is
             essential to allow monitoring of model training across BO iterations.
+        :param compile_args: Keyword arguments to pass to the ``compile`` method of the
+            Keras model (:class:`~tf.keras.Model`).
+            See https://keras.io/api/models/model_training_apis/#compile-method for a
+            list of possible arguments. The ``optimizer`` and ``metrics`` arguments
+            must not be included.
         :raise ValueError: If ``model`` has unsupported layers, ``num_rff_features`` is less than 0,
-            or if the ``optimizer`` is not of a supported type.
+            if the ``optimizer`` is not of a supported type, or `compile_args` contains
+            disallowed arguments.
         """
+        if compile_args is None:
+            compile_args = {}
+
+        if not {"optimizer", "metrics"}.isdisjoint(compile_args):
+            raise ValueError(
+                "optimizer and metrics arguments must not be included in compile_args."
+            )
+
         if isinstance(model, DeepGP):
             self._model_closure = None
         else:
@@ -111,7 +128,10 @@ class DeepGaussianProcess(
                 f"received {type(self.optimizer.optimizer)} instead."
             )
 
-        self.original_lr = self.optimizer.optimizer.lr.numpy()
+        if not isinstance(
+            self.optimizer.optimizer.lr, tf.keras.optimizers.schedules.LearningRateSchedule
+        ):
+            self.original_lr = self.optimizer.optimizer.lr.numpy()
 
         epochs = 400
 
@@ -147,7 +167,9 @@ class DeepGaussianProcess(
             dtype=tf.float64,
         )
         self._model_keras = model.as_training_model()
-        self._model_keras.compile(self.optimizer.optimizer, metrics=self.optimizer.metrics)
+        self._model_keras.compile(
+            optimizer=self.optimizer.optimizer, metrics=self.optimizer.metrics, **compile_args
+        )
         self._absolute_epochs = 0
         self._continuous_optimisation = continuous_optimisation
 
@@ -208,6 +230,10 @@ class DeepGaussianProcess(
             finally:
                 self._model_keras.history.model = history_model
 
+        # don't try to serialize any other copies of the history callback
+        if isinstance(state.get("_last_optimization_result"), keras.callbacks.History):
+            state["_last_optimization_result"] = ...
+
         return state
 
     def __setstate__(self, state: dict[str, Any]) -> None:
@@ -262,6 +288,10 @@ class DeepGaussianProcess(
                 model.set_weights(weights)
                 self._model_keras.history.set_model(model)
 
+        # recover optimization result if necessary (and possible)
+        if state.get("_last_optimization_result") is ...:
+            self._last_optimization_result = getattr(self._model_keras, "history")
+
     def __repr__(self) -> str:
         """"""
         return f"DeepGaussianProcess({self.model_gpflux!r}, {self.optimizer.optimizer!r})"
@@ -274,6 +304,7 @@ class DeepGaussianProcess(
     def model_keras(self) -> tf.keras.Model:
         return self._model_keras
 
+    @inherit_check_shapes
     def sample(self, query_points: TensorType, num_samples: int) -> TensorType:
         trajectory = self.trajectory_sampler().get_trajectory()
         expanded_query_points = tf.expand_dims(query_points, -2)  # [N, 1, D]
@@ -335,7 +366,7 @@ class DeepGaussianProcess(
 
             inputs = layer(inputs)
 
-    def optimize(self, dataset: Dataset) -> None:
+    def optimize(self, dataset: Dataset) -> keras.callbacks.History:
         """
         Optimize the model with the specified `dataset`.
         :param dataset: The data with which to optimize the `model`.
@@ -358,10 +389,16 @@ class DeepGaussianProcess(
         if self._continuous_optimisation:
             self._absolute_epochs = self._absolute_epochs + len(hist.history["loss"])
 
-        # Reset lr in case there was an lr schedule: a schedule will have change the learning rate,
-        # so that the next time we call `optimize` the starting learning rate would be different.
-        # Therefore we make sure the learning rate is set back to its initial value.
-        self.optimizer.optimizer.lr.assign(self.original_lr)
+        # Reset lr in case there was an lr schedule: a schedule will have changed the learning
+        # rate, so that the next time we call `optimize` the starting learning rate would be
+        # different. Therefore, we make sure the learning rate is set back to its initial value.
+        # However, this is not needed for `LearningRateSchedule` instances.
+        if not isinstance(
+            self.optimizer.optimizer.lr, tf.keras.optimizers.schedules.LearningRateSchedule
+        ):
+            self.optimizer.optimizer.lr.assign(self.original_lr)
+
+        return hist
 
     def log(self, dataset: Optional[Dataset] = None) -> None:
         """
