@@ -15,10 +15,11 @@
 from __future__ import annotations
 
 from abc import ABC, abstractmethod
-from typing import Callable, Generic, Optional, TypeVar
+from typing import Any, Callable, Generic, Optional, Sequence, TypeVar
 
 import gpflow
 import tensorflow as tf
+from check_shapes import check_shapes
 from typing_extensions import Protocol, runtime_checkable
 
 from ..data import Dataset
@@ -45,6 +46,11 @@ class ProbabilisticModel(Protocol):
     """
 
     @abstractmethod
+    @check_shapes(
+        "query_points: [batch..., D]",
+        "return[0]: [batch..., E...]",
+        "return[1]: [batch..., E...]",
+    )
     def predict(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
         """
         Return the mean and variance of the independent marginal distributions at each point in
@@ -62,6 +68,10 @@ class ProbabilisticModel(Protocol):
         raise NotImplementedError
 
     @abstractmethod
+    @check_shapes(
+        "query_points: [batch..., N, D]",
+        "return: [batch..., S, N, E...]",
+    )
     def sample(self, query_points: TensorType, num_samples: int) -> TensorType:
         """
         Return ``num_samples`` samples from the independent marginal distributions at
@@ -74,30 +84,14 @@ class ProbabilisticModel(Protocol):
         """
         raise NotImplementedError
 
-    def predict_y(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
-        """
-        Return the mean and variance of the independent marginal distributions at each point in
-        ``query_points`` for the observations, including noise contributions.
-
-        Note that this is not supported by all models.
-
-        :param query_points: The points at which to make predictions, of shape [..., D].
-        :return: The mean and variance of the independent marginal distributions at each point in
-            ``query_points``. For a predictive distribution with event shape E, the mean and
-            variance will both have shape [...] + E.
-        """
-        pass  # (required so that mypy doesn't think this method is abstract)
-        raise NotImplementedError(
-            f"Model {self!r} does not support predicting observations, just the latent function"
-        )
-
+    @abstractmethod
     def log(self, dataset: Optional[Dataset] = None) -> None:
         """
         Log model-specific information at a given optimization step.
 
         :param dataset: Optional data that can be used to log additional data-based model summaries.
         """
-        return
+        raise NotImplementedError
 
 
 @runtime_checkable
@@ -114,12 +108,13 @@ class TrainableProbabilisticModel(ProbabilisticModel, Protocol):
         raise NotImplementedError
 
     @abstractmethod
-    def optimize(self, dataset: Dataset) -> None:
+    def optimize(self, dataset: Dataset) -> Any:
         """
         Optimize the model objective with respect to (hyper)parameters given the specified
         ``dataset``.
 
         :param dataset: The data with which to train the model.
+        :return: Any (optimizer-specific) optimization result.
         """
         raise NotImplementedError
 
@@ -129,6 +124,11 @@ class SupportsPredictJoint(ProbabilisticModel, Protocol):
     """A probabilistic model that supports predict_joint."""
 
     @abstractmethod
+    @check_shapes(
+        "query_points: [batch..., B, D]",
+        "return[0]: [batch..., B, E...]",
+        "return[1]: [batch..., E..., B, B]",
+    )
     def predict_joint(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
         """
         :param query_points: The points at which to make predictions, of shape [..., B, D].
@@ -139,6 +139,29 @@ class SupportsPredictJoint(ProbabilisticModel, Protocol):
         raise NotImplementedError
 
 
+@runtime_checkable
+class SupportsPredictY(ProbabilisticModel, Protocol):
+    """A probabilistic model that supports predict_y."""
+
+    @abstractmethod
+    @check_shapes(
+        "query_points: [broadcast batch..., D]",
+        "return[0]: [batch..., E...]",
+        "return[1]: [batch..., E...]",
+    )
+    def predict_y(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
+        """
+        Return the mean and variance of the independent marginal distributions at each point in
+        ``query_points`` for the observations, including noise contributions.
+
+        :param query_points: The points at which to make predictions, of shape [..., D].
+        :return: The mean and variance of the independent marginal distributions at each point in
+            ``query_points``. For a predictive distribution with event shape E, the mean and
+            variance will both have shape [...] + E.
+        """
+        raise NotImplementedError
+
+
 @runtime_checkable
 class SupportsGetKernel(ProbabilisticModel, Protocol):
     """A probabilistic model that supports get_kernel."""
@@ -152,6 +175,11 @@ class SupportsGetKernel(ProbabilisticModel, Protocol):
         raise NotImplementedError
 
 
+@runtime_checkable
+class TrainableSupportsGetKernel(TrainableProbabilisticModel, SupportsGetKernel, Protocol):
+    """A trainable probabilistic model that supports get_kernel."""
+
+
 @runtime_checkable
 class SupportsGetObservationNoise(ProbabilisticModel, Protocol):
     """A probabilistic model that supports get_observation_noise."""
@@ -304,8 +332,6 @@ class SupportsReparamSamplerObservationNoise(
 ):
     """A model that supports both reparam_sampler and get_observation_noise."""
 
-    pass
-
 
 class ModelStack(ProbabilisticModel, Generic[ProbabilisticModelType]):
     r"""
@@ -338,6 +364,9 @@ class ModelStack(ProbabilisticModel, Generic[ProbabilisticModelType]):
         """
         self._models, self._event_sizes = zip(*(model_with_event_size,) + models_with_event_sizes)
 
+    # NB we don't use @inherit_shapes below as some classes break the shape API (👀 fantasizer)
+    # instead we rely on the shape checking inside the submodels
+
     def predict(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
         r"""
         :param query_points: The points at which to make predictions, of shape [..., D].
@@ -360,18 +389,6 @@ class ModelStack(ProbabilisticModel, Generic[ProbabilisticModelType]):
         samples = [model.sample(query_points, num_samples) for model in self._models]
         return tf.concat(samples, axis=-1)
 
-    def predict_y(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
-        r"""
-        :param query_points: The points at which to make predictions, of shape [..., D].
-        :return: The predictions from all the wrapped models, concatenated along the event axis in
-            the same order as they appear in :meth:`__init__`. If the wrapped models have predictive
-            distributions with event shapes [:math:`E_i`], the mean and variance will both have
-            shape [..., :math:`\sum_i E_i`].
-        :raise NotImplementedError: If any of the models don't implement predict_y.
-        """
-        means, vars_ = zip(*[model.predict_y(query_points) for model in self._models])
-        return tf.concat(means, axis=-1), tf.concat(vars_, axis=-1)
-
     def log(self, dataset: Optional[Dataset] = None) -> None:
         """
         Log model-specific information at a given optimization step.
@@ -407,7 +424,7 @@ class TrainableModelStack(ModelStack[TrainableProbabilisticModel], TrainableProb
         for model, obs in zip(self._models, observations):
             model.update(Dataset(dataset.query_points, obs))
 
-    def optimize(self, dataset: Dataset) -> None:
+    def optimize(self, dataset: Dataset) -> Sequence[Any]:
         """
         Optimize all the wrapped models on their corresponding data. The data for each model is
         extracted by splitting the observations in ``dataset`` along the event axis according to the
@@ -416,9 +433,12 @@ class TrainableModelStack(ModelStack[TrainableProbabilisticModel], TrainableProb
         :param dataset: The query points and observations for *all* the wrapped models.
         """
         observations = tf.split(dataset.observations, self._event_sizes, axis=-1)
+        results = []
 
         for model, obs in zip(self._models, observations):
-            model.optimize(Dataset(dataset.query_points, obs))
+            results.append(model.optimize(Dataset(dataset.query_points, obs)))
+
+        return results
 
 
 class HasReparamSamplerModelStack(ModelStack[HasReparamSampler], HasReparamSampler):
@@ -478,19 +498,55 @@ class PredictJointModelStack(ModelStack[SupportsPredictJoint], SupportsPredictJo
         return tf.concat(means, axis=-1), tf.concat(covs, axis=-3)
 
 
+class PredictYModelStack(ModelStack[SupportsPredictY], SupportsPredictY):
+    r"""
+    A :class:`PredictYModelStack` is a wrapper around a number of
+    :class:`SupportsPredictY`\ s.
+    It delegates :meth:`predict_y` to each model.
+    """
+
+    def predict_y(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
+        r"""
+        :param query_points: The points at which to make predictions, of shape [..., D].
+        :return: The predictions from all the wrapped models, concatenated along the event axis in
+            the same order as they appear in :meth:`__init__`. If the wrapped models have predictive
+            distributions with event shapes [:math:`E_i`], the mean and variance will both have
+            shape [..., :math:`\sum_i E_i`].
+        :raise NotImplementedError: If any of the models don't implement predict_y.
+        """
+        means, vars_ = zip(*[model.predict_y(query_points) for model in self._models])
+        return tf.concat(means, axis=-1), tf.concat(vars_, axis=-1)
+
+
 # It's useful, though a bit ugly, to define the stack constructors for some model type combinations
 class TrainableSupportsPredictJoint(TrainableProbabilisticModel, SupportsPredictJoint, Protocol):
     """A model that is both trainable and supports predict_joint."""
 
-    pass
-
 
 class TrainablePredictJointModelStack(
     TrainableModelStack, PredictJointModelStack, ModelStack[TrainableSupportsPredictJoint]
 ):
     """A stack of models that are both trainable and support predict_joint."""
 
-    pass
+
+class TrainableSupportsPredictY(TrainableProbabilisticModel, SupportsPredictY, Protocol):
+    """A model that is both trainable and supports predict_y."""
+
+
+class TrainablePredictYModelStack(
+    TrainableModelStack, PredictYModelStack, ModelStack[TrainableSupportsPredictY]
+):
+    """A stack of models that are both trainable and support predict_y."""
+
+
+class SupportsPredictJointPredictY(SupportsPredictJoint, SupportsPredictY, Protocol):
+    """A model that supports both predict_joint and predict_y."""
+
+
+class PredictJointPredictYModelStack(
+    PredictJointModelStack, PredictYModelStack, ModelStack[SupportsPredictJointPredictY]
+):
+    """A stack of models that support both predict_joint and predict_y."""
 
 
 class TrainableSupportsPredictJointHasReparamSampler(
@@ -498,8 +554,6 @@ class TrainableSupportsPredictJointHasReparamSampler(
 ):
     """A model that is trainable, supports predict_joint and has a reparameterization sampler."""
 
-    pass
-
 
 class TrainablePredictJointReparamModelStack(
     TrainablePredictJointModelStack,
@@ -508,8 +562,6 @@ class TrainablePredictJointReparamModelStack(
 ):
     """A stack of models that are both trainable and support predict_joint."""
 
-    pass
-
 
 class ReparametrizationSampler(ABC, Generic[ProbabilisticModelType]):
     r"""
