@@ -65,7 +65,7 @@ def build_keras_ensemble(
     input_tensor_spec, output_tensor_spec = get_tensor_spec_from_data(data)
 
     hidden_layer_args = []
-    for i in range(num_hidden_layers):
+    for _ in range(num_hidden_layers):
         hidden_layer_args.append({"units": units, "activation": activation})
 
     networks = [
