@@ -19,8 +19,10 @@ from typing import Optional
 
 import tensorflow as tf
 import tensorflow_probability as tfp
+from check_shapes import inherit_check_shapes
 from typing_extensions import Protocol, runtime_checkable
 
+from ...data import Dataset
 from ...types import TensorType
 from ..interfaces import ProbabilisticModel
 from ..optimizer import KerasOptimizer
@@ -60,9 +62,11 @@ class KerasPredictor(ProbabilisticModel, ABC):
         """The optimizer wrapper for training the model."""
         return self._optimizer
 
+    @inherit_check_shapes
     def predict(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
         return self.model.predict(query_points)
 
+    @inherit_check_shapes
     def sample(self, query_points: TensorType, num_samples: int) -> TensorType:
         raise NotImplementedError(
             """
@@ -73,6 +77,9 @@ class KerasPredictor(ProbabilisticModel, ABC):
             """
         )
 
+    def log(self, dataset: Optional[Dataset] = None) -> None:
+        return
+
 
 @runtime_checkable
 class DeepEnsembleModel(ProbabilisticModel, Protocol):
@@ -113,3 +120,8 @@ class DeepEnsembleModel(ProbabilisticModel, Protocol):
             of the ensemble.
         """
         raise NotImplementedError
+
+    @property
+    @abstractmethod
+    def dtype(self) -> tf.DType:
+        """The prediction dtype."""
