@@ -15,17 +15,20 @@
 from __future__ import annotations
 
 import re
-from typing import Any, Dict, Optional
+from typing import Any, Dict, Mapping, Optional
 
 import dill
+import keras.callbacks
 import tensorflow as tf
 import tensorflow_probability as tfp
 import tensorflow_probability.python.distributions as tfd
+from check_shapes import inherit_check_shapes
 from tensorflow.python.keras.callbacks import Callback
 
 from ... import logging
 from ...data import Dataset
 from ...types import TensorType
+from ...utils import flatten_leading_dims
 from ..interfaces import HasTrajectorySampler, TrainableProbabilisticModel, TrajectorySampler
 from ..optimizer import KerasOptimizer
 from ..utils import write_summary_data_based_metrics
@@ -82,6 +85,7 @@ class DeepEnsemble(
         bootstrap: bool = False,
         diversify: bool = False,
         continuous_optimisation: bool = True,
+        compile_args: Optional[Mapping[str, Any]] = None,
     ) -> None:
         """
         :param model: A Keras ensemble model with probabilistic networks as ensemble members. The
@@ -103,15 +107,28 @@ class DeepEnsemble(
         :param continuous_optimisation: If True (default), the optimizer will keep track of the
             number of epochs across BO iterations and use this number as initial_epoch. This is
             essential to allow monitoring of model training across BO iterations.
+        :param compile_args: Keyword arguments to pass to the ``compile`` method of the
+            Keras model (:class:`~tf.keras.Model`).
+            See https://keras.io/api/models/model_training_apis/#compile-method for a
+            list of possible arguments. The ``optimizer``, ``loss`` and ``metrics`` arguments
+            must not be included.
         :raise ValueError: If ``model`` is not an instance of
-            :class:`~trieste.models.keras.KerasEnsemble` or ensemble has less than two base
-            learners (networks).
+            :class:`~trieste.models.keras.KerasEnsemble`, or ensemble has less than two base
+            learners (networks), or `compile_args` contains disallowed arguments.
         """
         if model.ensemble_size < 2:
             raise ValueError(f"Ensemble size must be greater than 1 but got {model.ensemble_size}.")
 
         super().__init__(optimizer)
 
+        if compile_args is None:
+            compile_args = {}
+
+        if not {"optimizer", "loss", "metrics"}.isdisjoint(compile_args):
+            raise ValueError(
+                "optimizer, loss and metrics arguments must not be included in compile_args."
+            )
+
         if not self.optimizer.fit_args:
             self.optimizer.fit_args = {
                 "verbose": 0,
@@ -131,11 +148,16 @@ class DeepEnsemble(
             self.optimizer.metrics = ["mse"]
 
         model.model.compile(
-            self.optimizer.optimizer,
+            optimizer=self.optimizer.optimizer,
             loss=[self.optimizer.loss] * model.ensemble_size,
             metrics=[self.optimizer.metrics] * model.ensemble_size,
+            **compile_args,
         )
-        self.original_lr = self.optimizer.optimizer.lr.numpy()
+
+        if not isinstance(
+            self.optimizer.optimizer.lr, tf.keras.optimizers.schedules.LearningRateSchedule
+        ):
+            self.original_lr = self.optimizer.optimizer.lr.numpy()
         self._absolute_epochs = 0
         self._continuous_optimisation = continuous_optimisation
 
@@ -152,7 +174,7 @@ class DeepEnsemble(
 
     @property
     def model(self) -> tf.keras.Model:
-        """ " Returns compiled Keras ensemble model."""
+        """Returns compiled Keras ensemble model."""
         return self._model.model
 
     @property
@@ -222,6 +244,7 @@ class DeepEnsemble(
         x_transformed: dict[str, TensorType] = self.prepare_query_points(query_points)
         return self._model.model(x_transformed)
 
+    @inherit_check_shapes
     def predict(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
         r"""
         Returns mean and variance at ``query_points`` for the whole ensemble.
@@ -249,7 +272,11 @@ class DeepEnsemble(
         :return: The predicted mean and variance of the observations at the specified
             ``query_points``.
         """
-        ensemble_distributions = self.ensemble_distributions(query_points)
+        # handle leading batch dimensions, while still allowing `Functional` to
+        # "allow (None,) and (None, 1) Tensors to be passed interchangeably"
+        input_dims = min(len(query_points.shape), len(self.model.input_shape[0]))
+        flat_x, unflatten = flatten_leading_dims(query_points, output_dims=input_dims)
+        ensemble_distributions = self.ensemble_distributions(flat_x)
         predicted_means = tf.math.reduce_mean(
             [dist.mean() for dist in ensemble_distributions], axis=0
         )
@@ -260,7 +287,12 @@ class DeepEnsemble(
             - predicted_means**2
         )
 
-        return predicted_means, predicted_vars
+        return unflatten(predicted_means), unflatten(predicted_vars)
+
+    @property
+    def dtype(self) -> tf.DType:
+        """The prediction dtype."""
+        return self._model.output_dtype
 
     def predict_ensemble(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
         """
@@ -282,6 +314,7 @@ class DeepEnsemble(
 
         return predicted_means, predicted_vars
 
+    @inherit_check_shapes
     def sample(self, query_points: TensorType, num_samples: int) -> TensorType:
         """
         Return ``num_samples`` samples at ``query_points``. We use the mixture approximation in
@@ -340,7 +373,7 @@ class DeepEnsemble(
         """
         return
 
-    def optimize(self, dataset: Dataset) -> None:
+    def optimize(self, dataset: Dataset) -> keras.callbacks.History:
         """
         Optimize the underlying Keras ensemble model with the specified ``dataset``.
 
@@ -375,10 +408,16 @@ class DeepEnsemble(
         if self._continuous_optimisation:
             self._absolute_epochs = self._absolute_epochs + len(history.history["loss"])
 
-        # Reset lr in case there was an lr schedule: a schedule will have change the learning rate,
-        # so that the next time we call `optimize` the starting learning rate would be different.
-        # Therefore, we make sure the learning rate is set back to its initial value.
-        self.optimizer.optimizer.lr.assign(self.original_lr)
+        # Reset lr in case there was an lr schedule: a schedule will have changed the learning
+        # rate, so that the next time we call `optimize` the starting learning rate would be
+        # different. Therefore, we make sure the learning rate is set back to its initial value.
+        # However, this is not needed for `LearningRateSchedule` instances.
+        if not isinstance(
+            self.optimizer.optimizer.lr, tf.keras.optimizers.schedules.LearningRateSchedule
+        ):
+            self.optimizer.optimizer.lr.assign(self.original_lr)
+
+        return history
 
     def log(self, dataset: Optional[Dataset] = None) -> None:
         """
@@ -418,16 +457,15 @@ class DeepEnsemble(
                     else:
                         # unrecognised history key; ignore
                         continue
-                    if "model" in post and not logging.include_summary("_ensemble"):
-                        break
-                    else:
-                        if "model" in post:
-                            pre = pre + "/_ensemble"
-                        logging.histogram(f"{pre}/epoch{post}", lambda: v)
-                        logging.scalar(f"{pre}/final{post}", lambda: v[-1])
-                        logging.scalar(f"{pre}/diff{post}", lambda: v[0] - v[-1])
-                        logging.scalar(f"{pre}/min{post}", lambda: tf.reduce_min(v))
-                        logging.scalar(f"{pre}/max{post}", lambda: tf.reduce_max(v))
+                    if "model" in post:
+                        if not logging.include_summary("_ensemble"):
+                            break
+                        pre = pre + "/_ensemble"
+                    logging.histogram(f"{pre}/epoch{post}", lambda: v)
+                    logging.scalar(f"{pre}/final{post}", lambda: v[-1])
+                    logging.scalar(f"{pre}/diff{post}", lambda: v[0] - v[-1])
+                    logging.scalar(f"{pre}/min{post}", lambda: tf.reduce_min(v))
+                    logging.scalar(f"{pre}/max{post}", lambda: tf.reduce_max(v))
                 if dataset:
                     write_summary_data_based_metrics(
                         dataset=dataset, model=self, prefix="training_"
@@ -478,6 +516,11 @@ class DeepEnsemble(
                     tensorboard_writers,
                 ):
                     callback._writers = writers
+
+        # don't serialize any history optimization result
+        if isinstance(state.get("_last_optimization_result"), keras.callbacks.History):
+            state["_last_optimization_result"] = ...
+
         return state
 
     def __setstate__(self, state: dict[str, Any]) -> None:
@@ -504,3 +547,7 @@ class DeepEnsemble(
             loss=[self.optimizer.loss] * self._model.ensemble_size,
             metrics=[self.optimizer.metrics] * self._model.ensemble_size,
         )
+
+        # recover optimization result if necessary (and possible)
+        if state.get("_last_optimization_result") is ...:
+            self._last_optimization_result = getattr(self.model, "history")
