@@ -57,7 +57,7 @@ class DeepEnsembleTrajectorySampler(TrajectorySampler[DeepEnsembleModel]):
             raise NotImplementedError(
                 f"EnsembleTrajectorySampler only works with DeepEnsembleModel models, that support "
                 f"ensemble_size and ensemble_distributions methods; "
-                f"received {model.__repr__()}"
+                f"received {model!r}"
             )
 
         super().__init__(model)
@@ -135,7 +135,7 @@ class deep_ensemble_trajectory(TrajectoryFunctionClass):
 
         if self._diversify:
             self._eps = tf.Variable(
-                tf.zeros([0, 0], dtype=tf.float64), shape=[None, None], trainable=False
+                tf.zeros([0, 0], dtype=model.dtype), shape=[None, None], trainable=False
             )
         else:
             self._indices = tf.Variable(
@@ -168,9 +168,9 @@ class deep_ensemble_trajectory(TrajectoryFunctionClass):
 
         if self._diversify:
             predicted_means, predicted_vars = self._model.predict(flat_x)  # ([N*B, L], [N*B, L])
-            predicted_vars = predicted_vars + tf.cast(DEFAULTS.JITTER, predicted_vars.dtype)
+            predicted_vars = predicted_vars + tf.constant(DEFAULTS.JITTER, predicted_vars.dtype)
             predictions = predicted_means + tf.sqrt(predicted_vars) * tf.tile(
-                tf.cast(self._eps, predicted_vars.dtype), [tf.shape(x)[0], 1]
+                self._eps, [tf.shape(x)[0], 1]
             )  # [N*B, L]
             return unflatten(predictions)  # [N, B, L]
         else:
@@ -201,7 +201,7 @@ class deep_ensemble_trajectory(TrajectoryFunctionClass):
             self._eps.assign(
                 tf.random.normal(
                     shape=(self._batch_size, self._model.num_outputs),
-                    dtype=tf.float64,
+                    dtype=self._model.dtype,
                     seed=self._seed,
                 )
             )  # [B]
