@@ -31,6 +31,7 @@ from .single_objectives import (
     GramacyLee,
     Hartmann3,
     Hartmann6,
+    Levy8,
     LogarithmicGoldsteinPrice,
     Michalewicz2,
     Michalewicz5,
@@ -43,3 +44,4 @@ from .single_objectives import (
     SingleObjectiveTestProblem,
     Trid10,
 )
+from .utils import mk_multi_observer, mk_observer
